# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015, 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The ant plugin is useful for ant based parts.

The ant build system is commonly used to build Java projects.
The plugin requires a build.xml in the root of the source tree.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - ant-properties:
      (object)
      A dictionary of key-value pairs. Set the following properties when
      running ant.

    - ant-build-targets:
      (list of strings)
      Run the given ant targets.

    - ant-version:
      (string)
      The version of ant you want to use to build the source artifacts.
      Defaults to the current release downloadable from
      https://archive.apache.org/dist/ant/binaries/.

    - ant-version-checksum:
      (string)
      The checksum for ant-version in the form of <digest-type>/<digest>.
      As an example "sha512/2a803f578f341e164f6753e410413d16ab60fab...".

    - ant-openjdk-version:
      (string)
      openjdk version available to the base to use. If not set the latest
      version available to the base will be used.
"""

import logging
import os
from glob import glob
from typing import Sequence
from urllib.parse import urlsplit

import snapcraft
from snapcraft import formatting_utils
from snapcraft.internal import errors, sources


logger = logging.getLogger(__name__)


_DEFAULT_ANT_VERSION = "1.10.5"
_DEFAULT_ANT_CHECKSUM = "sha512/a7f1e0cec9d5ed1b3ab6cddbb9364f127305a997bbc88ecd734f9ef142ec0332375e01ace3592759bb5c3307cd9c1ac0a78a30053f304c7030ea459498e4ce4e"
_ANT_URL = (
    "https://archive.apache.org/dist/ant/binaries/apache-ant-{version}-bin.tar.bz2"
)


class UnsupportedJDKVersionError(errors.SnapcraftError):

    fmt = (
        "The ant-openjdk-version plugin property was set to {version!r}.\n"
        "Valid values for the {base!r} base are: {valid_versions}."
    )

    def __init__(
        self, *, base: str, version: str, valid_versions: Sequence[str]
    ) -> None:
        super().__init__(
            base=base,
            version=version,
            valid_versions=formatting_utils.humanize_list(
                valid_versions, conjunction="or"
            ),
        )


class AntPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()

        schema["properties"]["ant-properties"] = {"type": "object", "default": {}}

        schema["properties"]["ant-build-targets"] = {
            "type": "array",
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }

        schema["properties"]["ant-version"] = {"type": "string"}

        schema["properties"]["ant-version-checksum"] = {"type": "string"}

        schema["properties"]["ant-openjdk-version"] = {"type": "string", "default": ""}

        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the pull step dirty.
        return ["ant-version", "ant-version-checksum", "ant-openjdk-version"]

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return ["ant-build-targets", "ant-properties"]

    @property
    def _ant_tar(self):
        if self._ant_tar_handle is None:
            ant_uri = _ANT_URL.format(version=self._ant_version)
            self._ant_tar_handle = sources.Tar(
                ant_uri, self._ant_dir, source_checksum=self._ant_checksum
            )
        return self._ant_tar_handle

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        self._setup_ant()
        self._setup_base_tools(project.info.base)

    def _setup_base_tools(self, base):
        if base not in ("core16", "core18"):
            raise errors.PluginBaseError(
                part_name=self.name, base=self.project.info.base
            )

        if base == "core16":
            valid_versions = ["8", "9"]
        elif base == "core18":
            valid_versions = ["8", "11"]

        version = self.options.ant_openjdk_version
        if version and version not in valid_versions:
            raise UnsupportedJDKVersionError(
                version=version, base=base, valid_versions=valid_versions
            )
        elif not version:
            # Get the latest version from the slice
            version = valid_versions[-1]

        self.stage_packages.append("openjdk-{}-jre-headless".format(version))
        self.build_packages.append("openjdk-{}-jdk-headless".format(version))

    def _setup_ant(self):
        self._ant_tar_handle = None
        self._ant_dir = os.path.join(self.partdir, "ant")
        if self.options.ant_version:
            self._ant_version = self.options.ant_version
            self._ant_checksum = self.options.ant_version_checksum
        else:
            self._ant_version = _DEFAULT_ANT_VERSION
            self._ant_checksum = _DEFAULT_ANT_CHECKSUM

    def pull(self):
        super().pull()

        os.makedirs(self._ant_dir, exist_ok=True)
        self._ant_tar.download()

    def build(self):
        super().build()

        self._ant_tar.provision(self._ant_dir, clean_target=False, keep_tarball=True)

        command = ["ant"]

        if self.options.ant_build_targets:
            command.extend(self.options.ant_build_targets)

        for prop, value in self.options.ant_properties.items():
            command.extend(["-D{}={}".format(prop, value)])

        self.run(command, rootdir=self.builddir)
        files = glob(os.path.join(self.builddir, "target", "*.jar"))
        if files:
            jardir = os.path.join(self.installdir, "jar")
            os.makedirs(jardir)
            for f in files:
                base = os.path.basename(f)
                os.link(f, os.path.join(jardir, base))

        self._create_symlinks()

    def _create_symlinks(self):
        if self.project.info.base not in ("core18", "core19"):
            raise errors.PluginBaseError(
                part_name=self.name, base=self.project.info.base
            )

        os.makedirs(os.path.join(self.installdir, "bin"), exist_ok=True)
        java_bin = glob(
            os.path.join(
                self.installdir,
                "usr",
                "lib",
                "jvm",
                "java-{}-openjdk-*".format(self.options.ant_openjdk_version),
                "jre",
                "bin",
                "java",
            )
        )[0]
        os.symlink(
            os.path.relpath(java_bin, os.path.join(self.installdir, "bin")),
            os.path.join(self.installdir, "bin", "java"),
        )

    def run(self, cmd, rootdir):
        super().run(cmd, cwd=rootdir, env=self._build_environment())

    def get_proxy_options(self, scheme):
        proxy = os.environ.get("{}_proxy".format(scheme))
        if proxy:
            parsed = urlsplit(proxy)
            if parsed.hostname is not None:
                yield "-D{}.proxyHost={}".format(scheme, parsed.hostname)
            if parsed.port is not None:
                yield "-D{}.proxyPort={}".format(scheme, parsed.port)
            if parsed.username is not None:
                yield "-D{}.proxyUser={}".format(scheme, parsed.username)
            if parsed.password is not None:
                yield "-D{}.proxyPassword={}".format(scheme, parsed.password)

    def _build_environment(self):
        env = os.environ.copy()
        ant_bin = os.path.join(self._ant_dir, "bin")

        if env.get("PATH"):
            new_path = "{}:{}".format(ant_bin, env.get("PATH"))
        else:
            new_path = ant_bin

        env["PATH"] = new_path

        # Getting ant to use a proxy requires a little work; the JRE doesn't
        # help as much as it should.  (java.net.useSystemProxies=true ought
        # to do the trick, but it relies on desktop configuration rather
        # than using the standard environment variables.)
        ant_opts = []
        ant_opts.extend(self.get_proxy_options("http"))
        ant_opts.extend(self.get_proxy_options("https"))
        if ant_opts:
            env["ANT_OPTS"] = " ".join(opt.replace("'", "'\\''") for opt in ant_opts)

        return env

    def env(self, root):
        env = super().env(root)
        jars = glob(os.path.join(self.installdir, "jar", "*.jar"))
        if jars:
            jars = [
                os.path.join(root, "jar", os.path.basename(x)) for x in sorted(jars)
            ]
            env.extend(["CLASSPATH={}:$CLASSPATH".format(":".join(jars))])
        return env



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The qmake plugin is useful for building qmake-based parts.

These are projects that are built using .pro files.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - options:
      (list of strings)
      additional options to pass to the qmake invocation.
    - qt-version:
      (string; default: qt5)
      Version of Qt to use with qmake. Valid options are: qt4 and qt5
    - project-files:
      (list of strings)
      list of .pro files to pass to the qmake invocation.
"""

import os

import snapcraft
from snapcraft import common
from snapcraft.internal import errors


class QmakePlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["options"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["properties"]["qt-version"] = {
            "type": "string",
            "enum": ["qt4", "qt5"],
            "default": "qt5",
        }
        schema["properties"]["project-files"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }

        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the pull step dirty.
        return ["qt-version"]

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return ["options", "project-files"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        if project.info.base not in ("core16", "core18"):
            raise errors.PluginBaseError(part_name=self.name, base=project.info.base)

        self.build_packages.append("make")
        if self.options.qt_version == "qt5":
            self.build_packages.extend(["qt5-qmake", "qtbase5-dev"])
        elif self.options.qt_version == "qt4":
            self.build_packages.extend(["qt4-qmake", "libqt4-dev"])
        else:
            raise RuntimeError(
                "Unsupported Qt version: {!r}".format(self.options.qt_version)
            )

    def build(self):
        super().build()

        env = self._build_environment()

        sources = []
        if self.options.project_files:
            sourcedir = self.sourcedir
            source_subdir = getattr(self.options, "source_subdir", None)
            if source_subdir:
                sourcedir = os.path.join(sourcedir, source_subdir)
            sources = [
                os.path.join(sourcedir, project_file)
                for project_file in self.options.project_files
            ]

        self.run(
            ["qmake"] + self._extra_config() + self.options.options + sources, env=env
        )

        self.run(["make", "-j{}".format(self.parallel_build_count)], env=env)

        self.run(["make", "install", "INSTALL_ROOT=" + self.installdir], env=env)

    def _extra_config(self):
        extra_config = []

        for root in [self.installdir, self.project.stage_dir]:
            paths = common.get_library_paths(root, self.project.arch_triplet)
            for path in paths:
                extra_config.append('LIBS+="-L{}"'.format(path))

            paths = common.get_include_paths(root, self.project.arch_triplet)
            for path in paths:
                extra_config.append('INCLUDEPATH+="{}"'.format(path))

        return extra_config

    def _build_environment(self):
        env = os.environ.copy()
        env["QT_SELECT"] = self.options.qt_version

        return env



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import platform

import snapcraft


def _is_ci_env():
    env_prefixes = ["TRAVIS", "AUTOPKGTEST_TMP"]
    matches = []

    for prefix in env_prefixes:
        matches += [var for var in os.environ.keys() if var.startswith(prefix)]

    return len(matches) > 0


def get_user_agent():
    arch = snapcraft.ProjectOptions().deb_arch
    testing = "(testing) " if _is_ci_env() else ""
    return "snapcraft/{} {}{} ({})".format(
        snapcraft.__version__,
        testing,
        "/".join(platform.dist()[0:2]),  # i.e. Ubuntu/16.04
        arch,
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import shutil
import tarfile
import tempfile

from . import errors
from ._base import FileBase


class Tar(FileBase):
    def __init__(
        self,
        source,
        source_dir,
        source_tag=None,
        source_commit=None,
        source_branch=None,
        source_depth=None,
        source_checksum=None,
    ):
        super().__init__(
            source,
            source_dir,
            source_tag,
            source_commit,
            source_branch,
            source_depth,
            source_checksum,
        )
        if source_tag:
            raise errors.SnapcraftSourceInvalidOptionError("tar", "source-tag")
        elif source_commit:
            raise errors.SnapcraftSourceInvalidOptionError("tar", "source-commit")
        elif source_branch:
            raise errors.SnapcraftSourceInvalidOptionError("tar", "source-branch")
        if source_depth:
            raise errors.SnapcraftSourceInvalidOptionError("tar", "source-depth")

    def provision(self, dst, clean_target=True, keep_tarball=False, src=None):
        # TODO add unit tests.
        if src:
            tarball = src
        else:
            tarball = os.path.join(self.source_dir, os.path.basename(self.source))

        if clean_target:
            tmp_tarball = tempfile.NamedTemporaryFile().name
            shutil.move(tarball, tmp_tarball)
            shutil.rmtree(dst)
            os.makedirs(dst)
            shutil.move(tmp_tarball, tarball)

        self._extract(tarball, dst)

        if not keep_tarball:
            os.remove(tarball)

    def _extract(self, tarball, dst):
        with tarfile.open(tarball) as tar:

            def filter_members(tar):
                """Filters members and member names:
                    - strips common prefix
                    - bans dangerous names"""
                members = tar.getmembers()
                common = os.path.commonprefix([m.name for m in members])

                # commonprefix() works a character at a time and will
                # consider "d/ab" and "d/abc" to have common prefix "d/ab";
                # check all members either start with common dir
                for m in members:
                    if not (
                        m.name.startswith(common + "/")
                        or m.isdir()
                        and m.name == common
                    ):
                        # commonprefix() didn't return a dir name; go up one
                        # level
                        common = os.path.dirname(common)
                        break

                for m in members:
                    if m.name == common:
                        continue
                    self._strip_prefix(common, m)
                    # We mask all files to be writable to be able to easily
                    # extract on top.
                    m.mode = m.mode | 0o200
                    yield m

            tar.extractall(members=filter_members(tar), path=dst)

    def _strip_prefix(self, common, member):
        if member.name.startswith(common + "/"):
            member.name = member.name[len(common + "/") :]
        # strip leading '/', './' or '../' as many times as needed
        member.name = re.sub(r"^(\.{0,2}/)*", r"", member.name)
        # do the same for linkname if this is a hardlink
        if member.islnk() and not member.issym():
            if member.linkname.startswith(common + "/"):
                member.linkname = member.linkname[len(common + "/") :]
            member.linkname = re.sub(r"^(\.{0,2}/)*", r"", member.linkname)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

import pymacaroons

from . import errors

logger = logging.getLogger(__name__)


def _macaroon_auth(conf):
    """Format a macaroon and its associated discharge.

    :return: A string suitable to use in an Authorization header.

    """
    root_macaroon_raw = conf.get("macaroon")
    if root_macaroon_raw is None:
        raise errors.InvalidCredentialsError("Root macaroon not in the config file")
    unbound_raw = conf.get("unbound_discharge")
    if unbound_raw is None:
        raise errors.InvalidCredentialsError("Unbound discharge not in the config file")

    root_macaroon = _deserialize_macaroon(root_macaroon_raw)
    unbound = _deserialize_macaroon(unbound_raw)
    bound = root_macaroon.prepare_for_request(unbound)
    discharge_macaroon_raw = bound.serialize()
    auth = "Macaroon root={}, discharge={}".format(
        root_macaroon_raw, discharge_macaroon_raw
    )
    return auth


def _deserialize_macaroon(value):
    try:
        return pymacaroons.Macaroon.deserialize(value)
    except:  # noqa LP: #1733004
        raise errors.InvalidCredentialsError("Failed to deserialize macaroon")


from ._store_client import StoreClient  # noqa



import os
import urllib.parse

from ._client import Client

from . import constants


class UpDownClient(Client):
    """The Up/Down server provide upload/download snap capabilities."""

    def __init__(self, conf):
        super().__init__(
            conf,
            os.environ.get(
                "UBUNTU_STORE_UPLOAD_ROOT_URL", constants.UBUNTU_STORE_UPLOAD_ROOT_URL
            ),
        )

    def upload(self, monitor):
        return self.post(
            urllib.parse.urljoin(self.root_url, "unscanned-upload/"),
            data=monitor,
            headers={
                "Content-Type": monitor.content_type,
                "Accept": "application/json",
            },
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import signal
import shutil
import subprocess
from typing import Any, Callable, Dict, List, Optional, Sequence, Union  # noqa: F401

from snapcraft.internal.build_providers import errors


logger = logging.getLogger(__name__)


def _run(command: Sequence[str], stdin=subprocess.DEVNULL) -> None:
    logger.debug("Running {}".format(" ".join(command)))
    subprocess.check_call(command, stdin=stdin)


def _run_output(command: Sequence[str], **kwargs) -> bytes:
    logger.debug("Running {}".format(" ".join(command)))
    return subprocess.check_output(command, **kwargs)


def _ignore_signal(sig, stack):
    # We only do this for SIGUSR1 as it is the one we mostly
    # care about.
    if sig == signal.SIGUSR1:
        sig = "SIGUSR1"
    logger.debug("Ignoring {!r}: {!r}".format(sig, dir(stack)))


class MultipassCommand:
    """An object representation of common multipass cli commands."""

    provider_name = "multipass"

    def __init__(self) -> None:
        """Initialize a MultipassCommand instance.

        :raises errors.ProviderCommandNotFound:
            if the multipass command is not found.
        """
        provider_cmd = "multipass"
        if not shutil.which(provider_cmd):
            raise errors.ProviderCommandNotFound(command=provider_cmd)
        self.provider_cmd = provider_cmd
        # Workaround for https://github.com/CanonicalLtd/multipass/issues/221
        signal.signal(signal.SIGUSR1, _ignore_signal)

    def launch(
        self,
        *,
        instance_name: str,
        image: str,
        cpus: str = None,
        mem: str = None,
        disk: str = None,
        remote: str = None,
        cloud_init: str = None
    ) -> None:
        """Passthrough for running multipass launch.

        :param str instance_name: the name the launched instance will have.
        :param str image: the image to create the instance with.
        :param str cpus: amount of virtual CPUs to assign to the launched instance.
        :param str mem: amount of RAM to assign to the launched instance.
        :param str disk: amount of disk space the instance will see.
        :param str remote: the remote server to retrieve the image from.
        :param str cloud_init_file: path to a user-data cloud-init configuration.
        """
        if remote is not None:
            image = "{}:{}".format(remote, image)
        cmd = [self.provider_cmd, "launch", image, "--name", instance_name]
        if cloud_init is not None:
            cmd.extend(["--cloud-init", cloud_init])
        if cpus is not None:
            cmd.extend(["--cpus", cpus])
        if mem is not None:
            cmd.extend(["--mem", mem])
        if disk is not None:
            cmd.extend(["--disk", disk])
        try:
            _run(cmd)
        except subprocess.CalledProcessError as process_error:
            raise errors.ProviderLaunchError(
                provider_name=self.provider_name, exit_code=process_error.returncode
            ) from process_error

    def start(self, *, instance_name: str) -> None:
        """Passthrough for running multipass start.

        :param str instance_name: the name of the instance to start.
        """
        cmd = [self.provider_cmd, "start", instance_name]
        try:
            _run(cmd)
        except subprocess.CalledProcessError as process_error:
            raise errors.ProviderStartError(
                provider_name=self.provider_name, exit_code=process_error.returncode
            ) from process_error

    def stop(self, *, instance_name: str, time: int = None) -> None:
        """Passthrough for running multipass stop.

        :param str instance_name: the name of the instance to stop.
        :param str time: time from now, in minutes, to delay shutdown of the
                         instance.
        """
        cmd = [self.provider_cmd, "stop"]
        if time:
            cmd.extend(["--time", str(time)])
        cmd.append(instance_name)
        try:
            _run(cmd)
        except subprocess.CalledProcessError as process_error:
            raise errors.ProviderStopError(
                provider_name=self.provider_name, exit_code=process_error.returncode
            ) from process_error

    def delete(self, *, instance_name: str, purge=True) -> None:
        """Passthrough for running multipass delete.

        :param str instance_name: the name of the instance to delete.
        :param bool purge: if true, purge the instance's image after deleting.
        """
        cmd = [self.provider_cmd, "delete", instance_name]
        if purge:
            cmd.append("--purge")
        try:
            _run(cmd)
        except subprocess.CalledProcessError as process_error:
            raise errors.ProviderDeleteError(
                provider_name=self.provider_name, exit_code=process_error.returncode
            ) from process_error

    def execute(
        self, *, command: Sequence[str], instance_name: str, hide_output: bool = False
    ) -> Optional[bytes]:
        """Passthrough for running multipass exec.

        :param list command: the command to exectute on the instance.
        :param str instance_name: the name of the instance to execute command.
        """
        cmd = [self.provider_cmd, "exec", instance_name, "--"] + list(command)
        output = None
        try:
            if hide_output:
                output = _run_output(cmd)
            else:
                _run(cmd, stdin=None)
        except subprocess.CalledProcessError as process_error:
            raise errors.ProviderExecError(
                provider_name=self.provider_name,
                command=command,
                exit_code=process_error.returncode,
            ) from process_error

        return output

    def shell(self, *, instance_name: str) -> None:
        """Passthrough for running multipass shell.

        :param str instance_name: the name of the instance to execute command.
        """
        try:
            _run([self.provider_cmd, "shell", instance_name], stdin=None)
        except subprocess.CalledProcessError as process_error:
            raise errors.ProviderShellError(
                provider_name=self.provider_name, exit_code=process_error.returncode
            ) from process_error

    def mount(
        self,
        *,
        source: str,
        target: str,
        uid_map: Dict[str, str] = None,
        gid_map: Dict[str, str] = None
    ) -> None:
        """Passthrough for running multipass mount.

        :param str source: path to the local directory to mount.
        :param str target: mountpoint inside the instance in the form of
                           <instance-name>:path.
        :param dict uid_map: A mapping of user IDs for use in the mount of the form
                             <host-id> -> <instance-id>.
                             File and folder ownership will be mapped from
                             <host> to <instance-name> inside the instance.
        :param dict gid_map: A mapping of group IDs for use in the mount of the form
                             <host-id> -> <instance-id>.
                             File and folder ownership will be mapped from
                             <host> to <instance-name> inside the instance.
        :raises errors.ProviderMountError: when the mount operation fails.
        """
        cmd = [self.provider_cmd, "mount", source, target]
        if uid_map is None:
            uid_map = dict()
        for host_map, instance_map in uid_map.items():
            cmd.extend(["--uid-map", "{}:{}".format(host_map, instance_map)])
        if gid_map is None:
            gid_map = dict()
        for host_map, instance_map in gid_map.items():
            cmd.extend(["--gid-map", "{}:{}".format(host_map, instance_map)])
        try:
            _run(cmd)
        except subprocess.CalledProcessError as process_error:
            raise errors.ProviderMountError(
                provider_name=self.provider_name, exit_code=process_error.returncode
            ) from process_error

    def umount(self, *, mount: str) -> None:
        """Passthrough for running multipass mount.

        :param str mount: mountpoint inside the instance in the form of
                           <instance-name>:path to unmount.
        :raises errors.ProviderUMountError: when the unmount operation fails.
        """
        cmd = [self.provider_cmd, "umount", mount]
        try:
            _run(cmd)
        except subprocess.CalledProcessError as process_error:
            raise errors.ProviderUnMountError(
                provider_name=self.provider_name, exit_code=process_error.returncode
            ) from process_error

    def copy_files(self, *, source: str, destination: str) -> None:
        """Passthrough for running multipass copy-files.

        :param str source: the source file to copy, using syntax expected
                           by multipass.
        :param str destination: the destination of the copied file, using
                                syntax expected by multipass.
        """
        cmd = [self.provider_cmd, "copy-files", source, destination]
        try:
            _run(cmd)
        except subprocess.CalledProcessError as process_error:
            raise errors.ProviderFileCopyError(
                provider_name=self.provider_name, exit_code=process_error.returncode
            ) from process_error

    def info(self, *, instance_name: str, output_format: str = None) -> bytes:
        """Passthrough for running multipass info."""
        cmd = [self.provider_cmd, "info", instance_name]
        if output_format is not None:
            cmd.extend(["--format", output_format])
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        if process.returncode != 0:
            raise errors.ProviderInfoError(
                provider_name=self.provider_name,
                exit_code=process.returncode,
                stderr=stderr,
            )
        return stdout



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import glob

from . import errors


def get_python_command(python_major_version, *, stage_dir, install_dir):
    """Find the python command to use, preferring staged over the part.

    We prefer the staged python as opposed to the in-part python in order to
    support one part that supplies python, with another part built `after` it
    wanting to use its python.

    :param str python_major_version: The python major version to find (2 or 3)
    :param str stage_dir: Path to the staging area
    :param str install_dir: Path to the part's install area

    :return: Path to the python command that was found
    :rtype: str

    :raises MissingPythonCommandError: If no python could be found in the
                                       staging or part's install area.
    """
    python_command_name = "python{}".format(python_major_version)
    python_command = os.path.join("usr", "bin", python_command_name)
    staged_python = os.path.join(stage_dir, python_command)
    part_python = os.path.join(install_dir, python_command)

    if os.path.exists(staged_python):
        return staged_python
    elif os.path.exists(part_python):
        return part_python
    else:
        raise errors.MissingPythonCommandError(
            python_command_name, [stage_dir, install_dir]
        )


def get_python_headers(python_major_version, *, stage_dir):
    """Find the python headers to use, if any, preferring staged over the host.

    We want to make sure we use the headers from the staging area if available,
    or we may end up building for an older python version than the one we
    actually want to use.

    :param str python_major_version: The python version to find (2 or 3)
    :param str stage_dir: Path to the staging area

    :return: Path to the python headers that were found ('' if none)
    :rtype: str
    """
    python_command_name = "python{}".format(python_major_version)
    base_match = os.path.join("usr", "include", "{}*".format(python_command_name))
    staged_python = glob.glob(os.path.join(stage_dir, base_match))
    host_python = glob.glob(os.path.join(os.path.sep, base_match))

    if staged_python:
        return staged_python[0]
    elif host_python:
        return host_python[0]
    else:
        return ""


def get_python_home(python_major_version, *, stage_dir, install_dir):
    """Find the correct PYTHONHOME, preferring staged over the part.

    We prefer the staged python as opposed to the in-part python in order to
    support one part that supplies python, with another part built `after` it
    wanting to use its python.

    :param str python_major_version: The python major version to find (2 or 3)
    :param str stage_dir: Path to the staging area
    :param str install_dir: Path to the part's install area

    :return: Path to the PYTHONHOME that was found
    :rtype: str

    :raises MissingPythonCommandError: If no python could be found in the
                                       staging or part's install area.
    """
    python_command = get_python_command(
        python_major_version, stage_dir=stage_dir, install_dir=install_dir
    )
    if python_command.startswith(stage_dir):
        return os.path.join(stage_dir, "usr")
    else:
        return os.path.join(install_dir, "usr")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import snapcraft.internal.errors


class NoSuchFileError(snapcraft.internal.errors.SnapcraftError):

    fmt = (
        "Failed to find part that provided path: {path!r} does not "
        "exist.\n"
        "Check the file path and try again."
    )

    def __init__(self, path):
        super().__init__(path=path)


class SnapcraftInspectError(snapcraft.internal.errors.SnapcraftError):
    # Use a different exit code for these errors so the orchestrating snapcraft can
    # differentiate them.
    def get_exit_code(self):
        return 3


class ProvidesInvalidFilePathError(SnapcraftInspectError):

    fmt = (
        "Failed to find part that provides path: {path!r} is not in the "
        "staging or priming area.\n"
        "Ensure the path is in the staging or priming area and try again."
    )

    def __init__(self, path):
        super().__init__(path=path)


class UntrackedFileError(SnapcraftInspectError):

    fmt = "No known parts provided {path!r}. It may have been provided by a scriptlet."

    def __init__(self, path):
        super().__init__(path=path)


class NoStepsRunError(SnapcraftInspectError):
    fmt = "Failed to get latest step: no steps have run"



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Marius Gripsgard (mariogrip@ubuntu.com)
# Copyright (C) 2016-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This rust plugin is useful for building rust based parts.

Rust uses cargo to drive the build.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - rust-channel
      (string)
      select rust channel (stable, beta, nightly)
    - rust-revision
      (string)
      select rust version
    - rust-features
      (list of strings)
      Features used to build optional dependencies
"""

import collections
import logging
import os
import shutil
from contextlib import suppress

import snapcraft
from snapcraft import sources
from snapcraft import shell_utils
from snapcraft.internal import errors

_RUSTUP = "https://static.rust-lang.org/rustup.sh"
logger = logging.getLogger(__name__)


class RustPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["rust-channel"] = {"type": "string"}
        schema["properties"]["rust-revision"] = {"type": "string"}
        schema["properties"]["rust-features"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        return ["rust-revision", "rust-channel"]

    @classmethod
    def get_build_properties(cls):
        return ["rust-features"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        if project.info.base not in ("core16", "core18"):
            raise errors.PluginBaseError(part_name=self.name, base=project.info.base)

        self.build_packages.extend(["gcc", "git", "curl", "file"])
        self._rustpath = os.path.join(self.partdir, "rust")
        self._rustc = os.path.join(self._rustpath, "bin", "rustc")
        self._rustdoc = os.path.join(self._rustpath, "bin", "rustdoc")
        self._cargo = os.path.join(self._rustpath, "bin", "cargo")
        self._cargo_dir = os.path.join(self.builddir, ".cargo")
        self._cargo_config = os.path.join(self._cargo_dir, "config")
        self._rustlib = os.path.join(self._rustpath, "lib")
        self._rustup_get = sources.Script(_RUSTUP, self._rustpath)
        self._rustup = os.path.join(self._rustpath, "rustup.sh")
        self._manifest = collections.OrderedDict()

    def _test(self):
        if self.project.is_cross_compiling:
            logger.warning(
                "Skipping the test target run as this is a cross compilation."
            )
            return

        cmd = [self._cargo, "test", "-j{}".format(self.parallel_build_count)]
        if self.options.rust_features:
            cmd.append("--features")
            cmd.append(" ".join(self.options.rust_features))
        self.run(cmd, env=self._build_env())

    def build(self):
        super().build()

        self._write_cross_compile_config()

        self._test()

        cmd = [
            self._cargo,
            "install",
            "-j{}".format(self.parallel_build_count),
            "--root",
            self.installdir,
            "--path",
            self.builddir,
        ]
        if self.project.is_cross_compiling:
            cmd.extend(["--target", self._target])
        if self.options.rust_features:
            cmd.append("--features")
            cmd.append(" ".join(self.options.rust_features))
        self.run(cmd, env=self._build_env())
        self._record_manifest()

    def _write_cross_compile_config(self):
        if not self.project.is_cross_compiling:
            return

        if os.path.isfile(self._cargo_config):
            return

        # Cf. http://doc.crates.io/config.html
        os.makedirs(self._cargo_dir, exist_ok=True)
        with open(os.path.join(self._cargo_dir, "config"), "w") as f:
            f.write(
                """
                [target.{}]
                linker = "{}"
                """.format(
                    self._target, "{}-gcc".format(self.project.arch_triplet)
                )
            )

    def _record_manifest(self):
        self._manifest["rustup-version"] = self.run_output([self._rustup, "--version"])
        self._manifest["rustc-version"] = self.run_output([self._rustc, "--version"])
        self._manifest["cargo-version"] = self.run_output([self._cargo, "--version"])
        with suppress(FileNotFoundError, IsADirectoryError):
            with open(os.path.join(self.builddir, "Cargo.lock")) as lock_file:
                self._manifest["cargo-lock-contents"] = lock_file.read()

    def get_manifest(self):
        return self._manifest

    def enable_cross_compilation(self):
        # Cf. rustc --print target-list
        targets = {
            "armhf": "armv7-{}-{}eabihf",
            "arm64": "aarch64-{}-{}",
            "i386": "i686-{}-{}",
            "amd64": "x86_64-{}-{}",
            "ppc64el": "powerpc64le-{}-{}",
        }
        fmt = targets.get(self.project.deb_arch)
        if not fmt:
            raise NotImplementedError(
                "{!r} is not supported as a target architecture when "
                "cross-compiling with the rust plugin".format(self.project.deb_arch)
            )
        self._target = fmt.format("unknown-linux", "gnu")

    def _build_env(self):
        env = os.environ.copy()
        env.update(
            {
                "RUSTC": self._rustc,
                "RUSTDOC": self._rustdoc,
                "RUST_PATH": self._rustlib,
                "RUSTFLAGS": self._rustflags(),
            }
        )
        return env

    def _rustflags(self):
        ldflags = shell_utils.getenv("LDFLAGS")
        rustldflags = ""
        flags = {flag for flag in ldflags.split(" ") if flag}
        for flag in flags:
            rustldflags += "-C link-arg={} ".format(flag)
        return rustldflags.strip()

    def pull(self):
        super().pull()
        self._fetch_rust()
        self._fetch_deps()

    def clean_pull(self):
        super().clean_pull()

        with suppress(FileNotFoundError):
            shutil.rmtree(self._rustpath)

    def clean_build(self):
        super().clean_build()

        with suppress(FileNotFoundError):
            shutil.rmtree(self._cargo_dir)

    def _fetch_rust(self):
        options = []

        if self.options.rust_revision:
            options.append("--revision={}".format(self.options.rust_revision))

        if self.options.rust_channel:
            if self.options.rust_channel in ["stable", "beta", "nightly"]:
                options.append("--channel={}".format(self.options.rust_channel))
            else:
                raise errors.SnapcraftEnvironmentError(
                    "{} is not a valid rust channel".format(self.options.rust_channel)
                )
        os.makedirs(self._rustpath, exist_ok=True)
        self._rustup_get.download()
        cmd = [
            self._rustup,
            "--prefix={}".format(self._rustpath),
            "--disable-sudo",
            "--save",
        ] + options
        if self.project.is_cross_compiling:
            cmd.append("--with-target={}".format(self._target))
        self.run(cmd)

    def _fetch_deps(self):
        if self.options.source_subdir:
            sourcedir = os.path.join(self.sourcedir, self.options.source_subdir)
        else:
            sourcedir = self.sourcedir

        self.run(
            [
                self._cargo,
                "fetch",
                "--manifest-path",
                os.path.join(sourcedir, "Cargo.toml"),
            ],
            env=self._build_env(),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import functools
import logging
import os
import sys
from distutils import util

import click

import snapcraft
from snapcraft.internal import log
from .assertions import assertionscli
from .containers import containerscli
from .discovery import discoverycli
from .legacy import legacycli
from .lifecycle import lifecyclecli
from .store import storecli
from .inspect import inspectcli
from .help import helpcli
from .extensions import extensioncli
from .version import versioncli, SNAPCRAFT_VERSION_TEMPLATE
from .ci import cicli
from ._command_group import SnapcraftGroup
from ._options import add_build_options
from ._errors import exception_handler


command_groups = [
    storecli,
    cicli,
    assertionscli,
    containerscli,
    discoverycli,
    helpcli,
    legacycli,
    lifecyclecli,
    extensioncli,
    versioncli,
    inspectcli,
]


@click.group(cls=SnapcraftGroup, invoke_without_command=True)
@click.version_option(
    message=SNAPCRAFT_VERSION_TEMPLATE, version=snapcraft.__version__  # type: ignore
)
@click.pass_context
@add_build_options(hidden=True)
@click.option("--debug", "-d", is_flag=True)
def run(ctx, debug, catch_exceptions=False, **kwargs):
    """Snapcraft is a delightful packaging tool."""

    # Debugging snapcraft itself is not tied to debugging a snapcraft project.
    try:
        is_snapcraft_developer_debug = util.strtobool(
            os.getenv("SNAPCRAFT_ENABLE_DEVELOPER_DEBUG", "n")
        )
    except ValueError:
        is_snapcraft_developer_debug = False
    if is_snapcraft_developer_debug:
        log_level = logging.DEBUG
        click.echo(
            "Starting snapcraft {} from {}.".format(
                snapcraft.__version__, os.path.dirname(__file__)
            )
        )
    else:
        log_level = logging.INFO

    # Setup global exception handler (to be called for unhandled exceptions)
    sys.excepthook = functools.partial(
        exception_handler, debug=is_snapcraft_developer_debug
    )

    # In an ideal world, this logger setup would be replaced
    log.configure(log_level=log_level)
    # The default command
    if not ctx.invoked_subcommand:
        ctx.forward(lifecyclecli.commands["snap"])


# This would be much easier if they were subcommands
for command_group in command_groups:
    for command in command_group.commands:
        run.add_command(command_group.commands[command])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import importlib

import click

from ._options import get_project


_SUPPORTED_CI_SYSTEMS = ("travis",)


@click.group()
def cicli():
    pass


@cicli.command("enable-ci")
@click.argument(
    "ci-system", metavar="<ci-system>", type=click.Choice(_SUPPORTED_CI_SYSTEMS)
)
@click.option(
    "--refresh",
    is_flag=True,
    help=("Refresh the macaroon required to be able to push and release to channels"),
)
def enableci(ci_system, refresh):
    """Enable continuous-integration systems to build and release snaps."""
    module = importlib.import_module("snapcraft.integrations.{}".format(ci_system))

    project = get_project()
    if refresh:
        module.refresh(project)
    else:
        if click.confirm(module.__doc__):
            module.enable(project)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import contextlib
import glob
import logging
import os
import re
import shutil
import subprocess
import tempfile
from typing import Dict, FrozenSet, List, Set, Sequence, Tuple, Union  # noqa

import elftools.elf.elffile
from pkg_resources import parse_version

from snapcraft import file_utils
from snapcraft.internal import common, errors, repo


logger = logging.getLogger(__name__)


class NeededLibrary:
    """Represents an ELF library version."""

    def __init__(self, *, name: str) -> None:
        self.name = name
        self.versions = set()  # type: Set[str]

    def add_version(self, version: str) -> None:
        self.versions.add(version)


ElfArchitectureTuple = Tuple[str, str, str]
ElfDataTuple = Tuple[
    ElfArchitectureTuple, str, str, Dict[str, NeededLibrary], bool
]  # noqa: E501
SonameCacheDict = Dict[Tuple[ElfArchitectureTuple, str], str]


# Old pyelftools uses byte strings for section names.  Some data is
# also returned as bytes, which is handled below.
if parse_version(elftools.__version__) >= parse_version("0.24"):
    _DYNAMIC = ".dynamic"  # type: Union[str, bytes]
    _GNU_VERSION_R = ".gnu.version_r"  # type: Union[str, bytes]
    _INTERP = ".interp"  # type: Union[str, bytes]
else:
    _DYNAMIC = b".dynamic"
    _GNU_VERSION_R = b".gnu.version_r"
    _INTERP = b".interp"


class SonameCache:
    """A cache for sonames."""

    def __getitem__(self, key):
        return self._soname_paths[key]

    def __setitem__(self, key, item):
        # Initial API error checks
        if not isinstance(key, tuple):
            raise EnvironmentError(
                "The key for SonameCache has to be a (arch, soname) tuple."
            )
        if not isinstance(key[0], tuple) or len(key[0]) != 3:
            raise EnvironmentError(
                "The first element of the key needs to of type ElfArchitectureTuple."
            )
        if not isinstance(key[1], str):
            raise EnvironmentError(
                "The second element of the key needs to be "
                "of type str representing the soname."
            )
        self._soname_paths[key] = item

    def __contains__(self, key):
        return key in self._soname_paths

    def __init__(self):
        """Initialize a cache for sonames"""
        self._soname_paths = dict()  # type: SonameCacheDict

    def reset_except_root(self, root):
        """Reset the cache values that aren't contained within root."""
        new_soname_paths = dict()  # type: SonameCacheDict
        for key, value in self._soname_paths.items():
            if value is not None and value.startswith(root):
                new_soname_paths[key] = value

        self._soname_paths = new_soname_paths


class Library:
    """Represents the SONAME and path to the library."""

    def __init__(
        self,
        *,
        soname: str,
        path: str,
        root_path: str,
        core_base_path: str,
        arch: ElfArchitectureTuple,
        soname_cache: SonameCache
    ) -> None:
        self.soname = soname

        # We need to always look for the soname inside root first,
        # and after exhausting all options look in core_base_path.
        if path.startswith(root_path):
            self.path = path
        else:
            self.path = _crawl_for_path(
                soname=soname,
                root_path=root_path,
                core_base_path=core_base_path,
                arch=arch,
                soname_cache=soname_cache,
            )

        if not self.path and path.startswith(core_base_path):
            self.path = path

        # Required for libraries on the host and the fetching mechanism
        if not self.path:
            self.path = path

        # self.path has the correct resulting path.
        if self.path.startswith(core_base_path):
            self.in_base_snap = True
        else:
            self.in_base_snap = False

        logger.debug(
            "{soname} with original path {original_path} found on {path} in base: {in_base}".format(
                soname=soname,
                original_path=path,
                path=self.path,
                in_base=self.in_base_snap,
            )
        )


def _crawl_for_path(
    *,
    soname: str,
    root_path: str,
    core_base_path: str,
    arch: ElfArchitectureTuple,
    soname_cache: SonameCache
) -> str:
    # Speed things up and return what was already found once.
    if (arch, soname) in soname_cache:
        return soname_cache[arch, soname]

    logger.debug("Crawling to find soname {!r}".format(soname))
    for path in (root_path, core_base_path):
        if not os.path.exists(path):
            continue
        for root, directories, files in os.walk(path):
            for file_name in files:
                if file_name == soname:
                    file_path = os.path.join(root, file_name)
                    if ElfFile.is_elf(file_path):
                        # We found a match by name, anyway. Let's verify that
                        # the architecture is the one we want.
                        elf_file = ElfFile(path=file_path)
                        if elf_file.arch == arch:
                            soname_cache[arch, soname] = file_path
                            return file_path

    # If not found we cache it too
    soname_cache[arch, soname] = None
    return None


# Old versions of pyelftools return bytes rather than strings for
# certain APIs.  So we pass those values through this function to get
# a consistent result.
def _ensure_str(s):
    if isinstance(s, bytes):
        return s.decode("ascii")
    assert isinstance(s, str)
    return s


class ElfFile:
    """ElfFile represents and elf file on a path and its attributes."""

    @classmethod
    def is_elf(cls, path: str) -> bool:
        if not os.path.isfile(path):
            # ELF binaries are regular files
            return False
        with open(path, "rb") as bin_file:
            return bin_file.read(4) == b"\x7fELF"

    def __init__(self, *, path: str) -> None:
        """Initialize an ElfFile instance.

        :param str path: path to an elf_file within a snapcraft project.
        """
        self.path = path
        self.dependencies = set()  # type: Set[Library]
        elf_data = self._extract(path)
        self.arch = elf_data[0]
        self.interp = elf_data[1]
        self.soname = elf_data[2]
        self.needed = elf_data[3]
        self.execstack_set = elf_data[4]

    def _extract(self, path: str) -> ElfDataTuple:  # noqa: C901
        arch = None  # type: ElfArchitectureTuple
        interp = str()
        soname = str()
        libs = dict()
        execstack_set = False

        with open(path, "rb") as fp:
            elf = elftools.elf.elffile.ELFFile(fp)

            # A set of fields to identify the architecture of the ELF file:
            #  EI_CLASS: 32/64 bit (e.g. amd64 vs. x32)
            #  EI_DATA: byte orer (e.g. ppc64 vs. ppc64le)
            #  e_machine: instruction set (e.g. x86-64 vs. arm64)
            #
            # For amd64 binaries, this will evaluate to:
            #   ('ELFCLASS64', 'ELFDATA2LSB', 'EM_X86_64')
            arch = (
                elf.header.e_ident.EI_CLASS,
                elf.header.e_ident.EI_DATA,
                elf.header.e_machine,
            )

            # If we are processing a detached debug info file, these
            # sections will be present but empty.
            interp_section = elf.get_section_by_name(_INTERP)
            if (
                interp_section is not None
                and interp_section.header.sh_type != "SHT_NOBITS"
            ):
                interp = interp_section.data().rstrip(b"\x00").decode("ascii")

            dynamic_section = elf.get_section_by_name(_DYNAMIC)
            if (
                dynamic_section is not None
                and dynamic_section.header.sh_type != "SHT_NOBITS"
            ):
                for tag in dynamic_section.iter_tags("DT_NEEDED"):
                    needed = _ensure_str(tag.needed)
                    libs[needed] = NeededLibrary(name=needed)
                for tag in dynamic_section.iter_tags("DT_SONAME"):
                    soname = _ensure_str(tag.soname)

            verneed_section = elf.get_section_by_name(_GNU_VERSION_R)
            if (
                verneed_section is not None
                and verneed_section.header.sh_type != "SHT_NOBITS"
            ):
                for library, versions in verneed_section.iter_versions():
                    library_name = _ensure_str(library.name)
                    # If the ELF file only references weak symbols
                    # from a library, it may be absent from DT_NEEDED
                    # but still have an entry in .gnu.version_r for
                    # symbol versions.
                    if library_name not in libs:
                        continue
                    lib = libs[library_name]
                    for version in versions:
                        lib.add_version(_ensure_str(version.name))

            for segment in elf.iter_segments():
                if segment["p_type"] == "PT_GNU_STACK":
                    # p_flags holds the bit mask for this segment.
                    # See `man 5 elf`.
                    mode = segment["p_flags"]
                    if mode & elftools.elf.constants.P_FLAGS.PF_X:
                        execstack_set = True

        return arch, interp, soname, libs, execstack_set

    def is_linker_compatible(self, *, linker_version: str) -> bool:
        """Determines if linker will work given the required glibc version."""
        version_required = self.get_required_glibc()
        r = parse_version(version_required) <= parse_version(linker_version)
        logger.debug(
            "Checking if linker {!r} will work with "
            "GLIBC_{} required by {!r}: {!r}".format(
                linker_version, version_required, self.path, r
            )
        )
        return r

    def get_required_glibc(self) -> str:
        """Returns the required glibc version for this ELF file."""
        with contextlib.suppress(AttributeError):
            return self._required_glibc  # type: ignore

        version_required = ""
        for lib in self.needed.values():
            for version in lib.versions:
                if not version.startswith("GLIBC_"):
                    continue
                version = version[6:]
                if parse_version(version) > parse_version(version_required):
                    version_required = version

        self._required_glibc = version_required
        return version_required

    def load_dependencies(
        self, root_path: str, core_base_path: str, soname_cache: SonameCache = None
    ) -> Set[str]:
        """Load the set of libraries that are needed to satisfy elf's runtime.

        This may include libraries contained within the project.
        The object's .dependencies attribute is set after loading.

        :param str root_path: the root path to search for missing dependencies.
        :param str core_base_path: the core base path to search for missing
                                   dependencies.
        :param SonameCache soname_cache: a cache of previously search
                                         dependencies.
        :returns: a set of string with paths to the library dependencies of
                  elf.
        """
        if soname_cache is None:
            soname_cache = SonameCache()

        logger.debug("Getting dependencies for {!r}".format(self.path))
        ldd_out = []  # type: List[str]
        try:
            # ldd output sample:
            # /lib64/ld-linux-x86-64.so.2 (0x00007fb3c5298000)
            # libm.so.6 => /lib/x86_64-linux-gnu/libm.so.6 (0x00007fb3bef03000)
            ldd_out = common.run_output(["ldd", self.path]).split("\n")
        except subprocess.CalledProcessError:
            logger.warning(
                "Unable to determine library dependencies for {!r}".format(self.path)
            )
            return set()
        ldd_out_split = [l.split() for l in ldd_out]
        libs = set()
        for ldd_line in ldd_out_split:
            if len(ldd_line) > 2:
                libs.add(
                    Library(
                        soname=ldd_line[0],
                        path=ldd_line[2],
                        root_path=root_path,
                        core_base_path=core_base_path,
                        arch=self.arch,
                        soname_cache=soname_cache,
                    )
                )

        self.dependencies = libs

        # Return a set useful only for fetching libraries from the host
        library_paths = set()  # type: Set[str]
        for l in libs:
            if os.path.exists(l.path) and not l.in_base_snap:
                library_paths.add(l.path)
        return library_paths


class Patcher:
    """Patcher holds the necessary logic to patch elf files."""

    def __init__(
        self, *, dynamic_linker: str, root_path: str, preferred_patchelf_path=None
    ) -> None:
        """Create a Patcher instance.

        :param str dynamic_linker: the path to the dynamic linker to set the
                                   elf file to.
        :param str root_path: the base path for the snap to determine
                              if use of $ORIGIN is possible.
        :param str preferred_patchelf_path: patch the necessary elf_files with
                                        this patchelf.
        """
        self._dynamic_linker = dynamic_linker
        self._root_path = root_path

        if preferred_patchelf_path:
            self._patchelf_cmd = preferred_patchelf_path
        else:
            self._patchelf_cmd = file_utils.get_tool_path("patchelf")

        self._strip_cmd = file_utils.get_tool_path("strip")

    def patch(self, *, elf_file: ElfFile) -> None:
        """Patch elf_file with the Patcher instance configuration.

        If the ELF is executable, patch it to use the configured linker.
        If the ELF has dependencies (DT_NEEDED), set an rpath to them.

        :param ElfFile elf: a data object representing an elf file and its
                            relevant attributes.
        :raises snapcraft.internal.errors.PatcherError:
            raised when the elf_file cannot be patched.
        """
        patchelf_args = []
        if elf_file.interp:
            patchelf_args.extend(["--set-interpreter", self._dynamic_linker])
        if elf_file.dependencies:
            rpath = self._get_rpath(elf_file)
            # Due to https://github.com/NixOS/patchelf/issues/94 we need
            # to first clear the current rpath
            self._run_patchelf(
                patchelf_args=["--remove-rpath"], elf_file_path=elf_file.path
            )
            # Parameters:
            # --force-rpath: use RPATH instead of RUNPATH.
            # --shrink-rpath: will remove unneeded entries, with the
            #                 side effect of preferring host libraries
            #                 so we simply do not use it.
            # --set-rpath: set the RPATH to the colon separated argument.
            patchelf_args.extend(["--force-rpath", "--set-rpath", rpath])

        # no patchelf_args means there is nothing to do.
        if not patchelf_args:
            return

        self._run_patchelf(patchelf_args=patchelf_args, elf_file_path=elf_file.path)

    def _run_patchelf(self, *, patchelf_args: List[str], elf_file_path: str) -> None:
        try:
            return self._do_run_patchelf(
                patchelf_args=patchelf_args, elf_file_path=elf_file_path
            )
        except errors.PatcherError as patch_error:
            # This is needed for patchelf to properly work with
            # go binaries (LP: #1736861).
            # We do this here instead of the go plugin for two reasons, the
            # first being that we do not want to blindly remove the section,
            # only doing it when necessary, and the second, this logic
            # should eventually be removed once patchelf catches up.
            try:
                logger.warning(
                    "Failed to update {!r}. Retrying after stripping "
                    "the .note.go.buildid from the elf file.".format(elf_file_path)
                )
                subprocess.check_call(
                    [
                        self._strip_cmd,
                        "--remove-section",
                        ".note.go.buildid",
                        elf_file_path,
                    ]
                )
            except subprocess.CalledProcessError:
                logger.warning(
                    "Could not properly strip .note.go.buildid "
                    "from {!r}.".format(elf_file_path)
                )
                raise patch_error
            return self._do_run_patchelf(
                patchelf_args=patchelf_args, elf_file_path=elf_file_path
            )

    def _do_run_patchelf(self, *, patchelf_args: List[str], elf_file_path: str) -> None:
        # Run patchelf on a copy of the primed file and replace it
        # after it is successful. This allows us to break the potential
        # hard link created when migrating the file across the steps of
        # the part.
        with tempfile.NamedTemporaryFile() as temp_file:
            shutil.copy2(elf_file_path, temp_file.name)

            cmd = [self._patchelf_cmd] + patchelf_args + [temp_file.name]
            try:
                subprocess.check_call(cmd)
            # There is no need to catch FileNotFoundError as patchelf should be
            # bundled with snapcraft which means its lack of existence is a
            # "packager" error.
            except subprocess.CalledProcessError as call_error:
                patchelf_version = (
                    subprocess.check_output([self._patchelf_cmd, "--version"])
                    .decode()
                    .strip()
                )
                # 0.10 is the version where patching certain binaries will
                # work (currently known affected packages are mostly built
                # with go).
                if parse_version(patchelf_version) < parse_version("0.10"):
                    raise errors.PatcherNewerPatchelfError(
                        elf_file=elf_file_path,
                        process_exception=call_error,
                        patchelf_version=patchelf_version,
                    )
                else:
                    raise errors.PatcherGenericError(
                        elf_file=elf_file_path, process_exception=call_error
                    )

            # We unlink to break the potential hard link
            os.unlink(elf_file_path)
            shutil.copy2(temp_file.name, elf_file_path)

    def _get_existing_rpath(self, elf_file_path):
        output = subprocess.check_output(
            [self._patchelf_cmd, "--print-rpath", elf_file_path]
        )
        return output.decode().strip().split(":")

    def _get_rpath(self, elf_file) -> str:
        origin_rpaths = list()  # type: List[str]
        base_rpaths = set()  # type: Set[str]
        existing_rpaths = self._get_existing_rpath(elf_file.path)

        for dependency in elf_file.dependencies:
            if dependency.path:
                if dependency.in_base_snap:
                    base_rpaths.add(os.path.dirname(dependency.path))
                elif dependency.path.startswith(self._root_path):
                    rel_library_path = os.path.relpath(dependency.path, elf_file.path)
                    rel_library_path_dir = os.path.dirname(rel_library_path)
                    # return the dirname, with the first .. replace
                    # with $ORIGIN
                    origin_rpath = rel_library_path_dir.replace("..", "$ORIGIN", 1)
                    if origin_rpath not in origin_rpaths:
                        origin_rpaths.append(origin_rpath)

        if existing_rpaths:
            # Only keep those that mention origin and are not already in our
            # bundle.
            existing_rpaths = [
                r for r in existing_rpaths if "$ORIGIN" in r and r not in origin_rpaths
            ]
            origin_rpaths = existing_rpaths + origin_rpaths

        origin_paths = ":".join((r for r in origin_rpaths if r))
        core_base_rpaths = ":".join(base_rpaths)

        if origin_paths and core_base_rpaths:
            return "{}:{}".format(origin_paths, core_base_rpaths)
        elif origin_paths and not core_base_rpaths:
            return origin_paths
        else:
            return core_base_rpaths


def determine_ld_library_path(root: str) -> List[str]:
    """Determine additional library paths needed for the linker loader.

    This is a workaround until full library searching is implemented which
    works by searching for ld.so.conf in specific hard coded locations
    within root.

    :param root str: the root directory to search for specific ld.so.conf
                     entries.
    :returns: a list of strings of library paths where relevant libraries
              can be found within root.
    """
    # If more ld.so.conf files need to be supported, add them here.
    ld_config_globs = {"{}/usr/lib/*/mesa*/ld.so.conf".format(root)}

    ld_library_paths = []
    for this_glob in ld_config_globs:
        for ld_conf_file in glob.glob(this_glob):
            ld_library_paths.extend(_extract_ld_library_paths(ld_conf_file))

    return [root + path for path in ld_library_paths]


def _extract_ld_library_paths(ld_conf_file: str) -> List[str]:
    # From the ldconfig manpage, paths can be colon-, space-, tab-, newline-,
    # or comma-separated.
    path_delimiters = re.compile(r"[:\s,]")
    comments = re.compile(r"#.*$")

    paths = []
    with open(ld_conf_file, "r") as f:
        for line in f:
            # Remove comments from line
            line = comments.sub("", line).strip()

            if line:
                paths.extend(path_delimiters.split(line))

    return paths


_libraries = None


def get_elf_files(root: str, file_list: Sequence[str]) -> FrozenSet[ElfFile]:
    """Return a frozenset of elf files from file_list prepended with root.

    :param str root: the root directory from where the file_list is generated.
    :param file_list: a list of file in root.
    :returns: a frozentset of ElfFile objects.
    """
    elf_files = set()  # type: Set[ElfFile]

    for part_file in file_list:
        # Filter out object (*.o) files-- we only care about binaries.
        if part_file.endswith(".o"):
            continue

        # No need to crawl links-- the original should be here, too.
        path = os.path.join(root, part_file)  # type: str
        if os.path.islink(path):
            logger.debug("Skipped link {!r} while finding dependencies".format(path))
            continue
        # Finally, make sure this is actually an ELF file
        if ElfFile.is_elf(path):
            elf_file = ElfFile(path=path)
            # if we have dyn symbols we are dynamic
            if elf_file.needed:
                elf_files.add(elf_file)

    return frozenset(elf_files)


def _get_dynamic_linker(library_list: List[str]) -> str:
    """Return the dynamic linker from library_list."""
    regex = re.compile(r"(?P<dynamic_linker>ld-[\d.]+.so)$")

    for library in library_list:
        m = regex.search(os.path.basename(library))
        if m:
            return library

    raise RuntimeError(
        "The format for the linker should be of the form "
        "<root>/ld-<X>.<Y>.so. There are no matches for the "
        "current libc6 package"
    )


def find_linker(*, root_path: str, snap_base_path: str) -> str:
    """Find and return the dynamic linker that would be seen at runtime.

    :param str root_path: the root path of a snap tree.
    :param str snap_base_path: absolute path to the snap once installed to
                               setup proper rpaths.
    :returns: the path to the dynamic linker to use
    """
    # We assume the current system will satisfy the GLIBC requirement,
    # get the current libc6 libraries (which includes the linker)
    libc6_libraries_list = repo.Repo.get_package_libraries("libc6")

    # For security reasons, we do not want to automatically pull in
    # libraries but expect them to be consciously brought in by stage-packages
    # instead.
    libc6_libraries_paths = [
        os.path.join(root_path, l[1:]) for l in libc6_libraries_list
    ]

    dynamic_linker = _get_dynamic_linker(libc6_libraries_paths)

    # Get the path to the "would be" dynamic linker when this snap is
    # installed. Strip the root_path from the retrieved dynamic_linker
    # variables + the leading `/` so that os.path.join can perform the
    # proper join with snap_base_path.
    dynamic_linker_path = os.path.join(
        snap_base_path, dynamic_linker[len(root_path) + 1 :]
    )

    return dynamic_linker_path



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import click
import collections
import inspect
import sys
import tabulate

from ._options import get_project
from snapcraft.internal import project_loader
from snapcraft import yaml_utils


@click.group()
def extensioncli(**kwargs):
    pass


@extensioncli.command("list-extensions")
def list_extensions(**kwargs):
    """List available extensions.

    This command has an alias of `extensions`.
    """

    extension_names = project_loader.supported_extension_names()

    if not extension_names:
        click.echo("No extensions supported")
        return

    extensions = []
    for extension_name in sorted(extension_names):
        extension = project_loader.find_extension(extension_name)
        extension_info = collections.OrderedDict()
        extension_info["Extension name"] = extension_name
        extension_info["Supported bases"] = ", ".join(sorted(extension.supported_bases))
        extensions.append(extension_info)

    click.echo(tabulate.tabulate(extensions, headers="keys"))


@extensioncli.command()
@click.pass_context
@click.argument("name")
def extension(ctx, name, **kwargs):
    """Show contents of extension."""

    extension_cls = project_loader.find_extension(name)

    # Not using inspect.getdoc here since it'll fall back to the base class
    docstring = extension_cls.__doc__
    if not docstring:
        raise project_loader.errors.ExtensionMissingDocumentationError(name)

    formatter = ctx.make_formatter()
    formatter.write_text(inspect.cleandoc(docstring))
    click.echo(formatter.getvalue().rstrip("\n"))


@extensioncli.command("expand-extensions")
def expand_extensions(**kwargs):
    """Display snapcraft.yaml with all extensions applied."""

    project = get_project(**kwargs)
    yaml_with_extensions = project_loader.apply_extensions(
        project.info.get_raw_snapcraft()
    )

    # Loading the config applied all the extensions, so just dump it back out
    yaml_utils.dump(yaml_with_extensions, stream=sys.stdout)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import Sequence

from snapcraft import config
from snapcraft.internal import (
    common,
    errors,
    meta,
    pluginhandler,
    project_loader,
    repo,
    states,
    steps,
)
from ._status_cache import StatusCache


logger = logging.getLogger(__name__)


def execute(
    step: steps.Step,
    project_config: "project_loader._config.Config",
    part_names: Sequence[str] = None,
):
    """Execute until step in the lifecycle for part_names or all parts.

    Lifecycle execution will happen for each step iterating over all
    the available parts, if part_names is specified, only those parts
    will run.

    If one of the parts to execute has an after keyword, execution is
    forced until the stage step for such part. If part_names was provided
    and after is not in this set, an exception will be raised.

    :param str step: A valid step in the lifecycle: pull, build, prime or snap.
    :param project_config: Fully loaded project (old logic moving either to
                           Project or the PluginHandler).
    :param list part_names: A list of parts to execute the lifecycle on.
    :raises RuntimeError: If a prerequesite of the part needs to be staged
                          and such part is not in the list of parts to iterate
                          over.
    :returns: A dict with the snap name, version, type and architectures.
    """
    installed_packages = repo.Repo.install_build_packages(project_config.build_tools)
    if installed_packages is None:
        raise ValueError(
            "The repo backend is not returning the list of installed packages"
        )

    installed_snaps = repo.snaps.install_snaps(project_config.build_snaps)

    try:
        global_state = states.GlobalState.load(
            filepath=project_config.project._global_state_file
        )
    except FileNotFoundError:
        global_state = states.GlobalState()
    global_state.append_build_packages(installed_packages)
    global_state.append_build_snaps(installed_snaps)
    global_state.save(filepath=project_config.project._global_state_file)

    executor = _Executor(project_config)
    executor.run(step, part_names)
    if not executor.steps_were_run:
        logger.warn(
            "The requested action has already been taken. Consider\n"
            "specifying parts, or clean the steps you want to run again."
        )

    return {
        "name": project_config.data["name"],
        "version": project_config.data.get("version"),
        "arch": project_config.data["architectures"],
        "type": project_config.data.get("type", ""),
    }


def _replace_in_part(part):
    for key, value in part.plugin.options.__dict__.items():
        replacements = project_loader.environment_to_replacements(
            project_loader.snapcraft_part_environment(part)
        )

        value = project_loader.replace_attr(value, replacements)
        setattr(part.plugin.options, key, value)

    return part


class _Executor:
    def __init__(self, project_config):
        self.config = project_config
        self.project = project_config.project
        self.parts_config = project_config.parts
        self.steps_were_run = False

        self._cache = StatusCache(project_config)

    def run(self, step: steps.Step, part_names=None):
        if part_names:
            self.parts_config.validate(part_names)
            # self.config.all_parts is already ordered, let's not lose that
            # and keep using a list.
            parts = [p for p in self.config.all_parts if p.name in part_names]
            processed_part_names = part_names
        else:
            parts = self.config.all_parts
            processed_part_names = self.config.part_names

        with config.CLIConfig() as cli_config:
            for current_step in step.previous_steps() + [step]:
                if current_step == steps.STAGE:
                    # XXX check only for collisions on the parts that have
                    # already been built --elopio - 20170713
                    pluginhandler.check_for_collisions(self.config.all_parts)
                for part in parts:
                    self._handle_step(part_names, part, step, current_step, cli_config)

        self._create_meta(step, processed_part_names)

    def _handle_step(
        self,
        requested_part_names: Sequence[str],
        part: pluginhandler.PluginHandler,
        requested_step: steps.Step,
        current_step: steps.Step,
        cli_config,
    ) -> None:
        # If this step hasn't yet run, all we need to do is run it
        if not self._cache.has_step_run(part, current_step):
            getattr(self, "_run_{}".format(current_step.name))(part)
            return

        # Alright, this step has already run. In that case, a few different
        # things need to happen:

        # 1. By default, if a step has already run, don't run it again.
        #    However, automatically clean and re-run the step if all the
        #    following conditions apply:
        #
        #      1.1. The step is the exact step that was requested (not an
        #           earlier one).
        #      1.2. The part was explicitly specified.
        if (
            requested_part_names
            and current_step == requested_step
            and part.name in requested_part_names
        ):
            getattr(self, "_re{}".format(current_step.name))(part)
            return

        # 2. If a step has already run, it might be dirty, in which case we
        #    need to clean and run it again.
        dirty_report = self._cache.get_dirty_report(part, current_step)
        if dirty_report:
            self._handle_dirty(part, current_step, dirty_report, cli_config)
            return

        # 3. If a step has already run, it might be outdated, in which case we
        #    need to update it (without cleaning if possible).
        outdated_report = self._cache.get_outdated_report(part, current_step)
        if outdated_report:
            self._handle_outdated(part, current_step, outdated_report, cli_config)
            return

        # 4. The step has already run, and is up-to-date, no need to run it
        #    again.
        notify_part_progress(
            part, "Skipping {}".format(current_step.name), "(already ran)"
        )

    def _run_pull(self, part):
        self._run_step(step=steps.PULL, part=part, progress="Pulling")

    def _repull(self, part, hint=""):
        self._rerun_step(
            step=steps.PULL,
            part=part,
            progress="Cleaning later steps and re-pulling",
            hint=hint,
        )

    def _run_build(self, part):
        self._run_step(step=steps.BUILD, part=part, progress="Building")

    def _rebuild(self, part, hint=""):
        self._rerun_step(
            step=steps.BUILD,
            part=part,
            progress="Cleaning later steps and re-building",
            hint=hint,
        )

    def _run_stage(self, part):
        self._run_step(step=steps.STAGE, part=part, progress="Staging")

    def _restage(self, part, hint=""):
        self._rerun_step(
            step=steps.STAGE,
            part=part,
            progress="Cleaning later steps and re-staging",
            hint=hint,
        )

    def _run_prime(self, part):
        self._run_step(step=steps.PRIME, part=part, progress="Priming")

    def _reprime(self, part, hint=""):
        self._rerun_step(step=steps.PRIME, part=part, progress="Re-priming", hint=hint)

    def _prepare_step(self, *, step: steps.Step, part: pluginhandler.PluginHandler):
        common.reset_env()
        all_dependencies = self.parts_config.get_dependencies(part.name)

        # Filter dependencies down to only those that need to run the
        # prerequisite step
        prerequisite_step = steps.get_dependency_prerequisite_step(step)
        dependencies = {
            p
            for p in all_dependencies
            if self._cache.should_step_run(p, prerequisite_step)
        }

        if dependencies:
            dependency_names = {p.name for p in dependencies}
            # Dependencies need to go all the way to the prerequisite step to
            # be able to share the common assets that make them a dependency
            logger.info(
                "{!r} has dependencies that need to be {}d: {}".format(
                    part.name, prerequisite_step.name, " ".join(dependency_names)
                )
            )
            self.run(prerequisite_step, dependency_names)

        # Run the preparation function for this step (if implemented)
        preparation_function = getattr(part, "prepare_{}".format(step.name), None)
        if preparation_function:
            notify_part_progress(part, "Preparing to {}".format(step.name), debug=True)
            preparation_function()

        common.env = self.parts_config.build_env_for_part(part)
        common.env.extend(self.config.project_env())

        part = _replace_in_part(part)

    def _run_step(self, *, step: steps.Step, part, progress, hint=""):
        self._prepare_step(step=step, part=part)

        notify_part_progress(part, progress, hint)
        getattr(part, step.name)()

        # We know we just ran this step, so rather than check, manually twiddle
        # the cache
        self._complete_step(part, step)

    def _complete_step(self, part, step):
        self._cache.clear_step(part, step)
        self._cache.add_step_run(part, step)
        self.steps_were_run = True

    def _rerun_step(self, *, step: steps.Step, part, progress, hint=""):
        staged_state = self.config.get_project_state(steps.STAGE)
        primed_state = self.config.get_project_state(steps.PRIME)

        # First clean the step, then run it again
        part.clean(staged_state, primed_state, step)

        # Uncache this and later steps since we just cleaned them: their status
        # has changed
        for current_step in [step] + step.next_steps():
            self._cache.clear_step(part, current_step)

        self._run_step(step=step, part=part, progress=progress, hint=hint)

    def _create_meta(self, step: steps.Step, part_names: Sequence[str]) -> None:
        if step == steps.PRIME and part_names == self.config.part_names:
            meta.create_snap_packaging(self.config)

    def _handle_dirty(self, part, step, dirty_report, cli_config):
        dirty_action = cli_config.get_outdated_step_action()
        if not step.clean_if_dirty:
            if dirty_action == config.OutdatedStepAction.ERROR:
                raise errors.StepOutdatedError(
                    step=step, part=part.name, dirty_report=dirty_report
                )

        getattr(self, "_re{}".format(step.name))(
            part, hint="({})".format(dirty_report.get_summary())
        )

    def _handle_outdated(self, part, step, outdated_report, cli_config):
        dirty_action = cli_config.get_outdated_step_action()
        if not step.clean_if_dirty:
            if dirty_action == config.OutdatedStepAction.ERROR:
                raise errors.StepOutdatedError(
                    step=step, part=part.name, outdated_report=outdated_report
                )

        update_function = getattr(part, "update_{}".format(step.name), None)
        if update_function:
            self._prepare_step(step=step, part=part)
            notify_part_progress(
                part,
                "Updating {} step for".format(step.name),
                "({})".format(outdated_report.get_summary()),
            )
            update_function()

            # We know we just ran this step, so rather than check, manually
            # twiddle the cache
            self._complete_step(part, step)
        else:
            getattr(self, "_re{}".format(step.name))(
                part, "({})".format(outdated_report.get_summary())
            )


def notify_part_progress(part, progress, hint="", debug=False):
    if debug:
        logger.debug("%s %s %s", progress, part.name, hint)
    else:
        logger.info("%s %s %s", progress, part.name, hint)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import datetime
import enum
import logging
import os
import tempfile
from typing import Callable, Generator, List, Optional
from typing import Any, Dict  # noqa: F401

from . import errors
from snapcraft import storeapi, yaml_utils
from snapcraft.internal import repo

logger = logging.getLogger(__name__)

_STORE_ASSERTION = [
    "account-key",
    "public-key-sha3-384=BWDEoaqyr25nF5SNCvEv2v7QnM9QsfCc0PBMYD_i2NGSQ32EF2d4D0hqUel3m8ul",
]


class _SnapOp(enum.Enum):
    NOP = 0
    INJECT = 1
    INSTALL = 2
    REFRESH = 3


def _get_snap_channel(snap_name: str) -> str:
    """Returns the channel to use for snap_name."""
    if snap_name == "snapcraft":
        channel = os.getenv(
            "SNAPCRAFT_BUILD_ENVIRONMENT_CHANNEL_SNAPCRAFT", "latest/stable"
        )
        logger.warning(
            "SNAPCRAFT_BUILD_ENVIRONMENT_CHANNEL_SNAPCRAFT is set: installing "
            "snapcraft from {}".format(channel)
        )
    else:
        channel = "latest/stable"

    return channel


class _SnapManager:
    def __init__(
        self,
        *,
        snap_name: str,
        snap_dir: str,
        latest_revision: str,
        snap_arch: str,
        inject_from_host: bool = True
    ) -> None:
        self.snap_name = snap_name
        self._snap_dir = snap_dir
        self._snap_arch = snap_arch
        self._inject_from_host = inject_from_host

        self._latest_revision = latest_revision
        self.__required_operation = None  # type: Optional[_SnapOp]
        self.__repo = None  # type: Optional[repo.snaps.SnapPackage]
        self.__revision = None  # type: Optional[str]
        self.__install_cmd = None  # type: Optional[List[str]]

    def _get_snap_repo(self):
        if self.__repo is None:
            self.__repo = repo.snaps.SnapPackage(self.snap_name)
        return self.__repo

    def get_op(self) -> _SnapOp:
        if self.__required_operation is not None:
            return self.__required_operation

        # From the point of view of multiple architectures if the target host (this)
        # is different than that of where these snaps run from, then we always need to
        # install from the store
        if self._inject_from_host:
            # Get information from the host.
            host_snap_repo = self._get_snap_repo()
            try:
                host_snap_info = host_snap_repo.get_local_snap_info()
                is_installed = host_snap_repo.installed
            except repo.errors.SnapdConnectionError:
                # This maybe because we are in a docker instance or another OS.
                is_installed = False
        else:
            is_installed = False

        # The evaluations for the required operation is as follows:
        # - if the snap is not installed on the host (is_installed == False),
        #   and the snap is not installed in the build environment
        #   (_latest_revision is None), then a store install will take place.
        # - else if the snap is not installed on the host (is_installed == False),
        #   but the is previously installed revision in the build environment
        #   (_latest_revision is not None), then a store install will take place.
        # - else if the snap is installed on the host (is_installed == True),
        #   and the snap installed in the build environment (_latest_revision) matches
        #   the one on the host, no operation takes place.
        # - else if the snap is installed on the host (is_installed == True),
        #   and the snap installed in the build environment (_latest_revision) does not
        #   match the one on the host, then a snap injection from the host will take place.
        if not is_installed and self._latest_revision is None:
            op = _SnapOp.INSTALL
        elif not is_installed and self._latest_revision is not None:
            op = _SnapOp.REFRESH
        elif is_installed and self._latest_revision == host_snap_info["revision"]:
            op = _SnapOp.NOP
        elif is_installed and self._latest_revision != host_snap_info["revision"]:
            op = _SnapOp.INJECT
        else:
            # This is a programmatic error
            raise RuntimeError(
                "Unhandled scenario for {!r} (host installed: {}, latest_revision {})".format(
                    self.snap_name, is_installed, self._latest_revision
                )
            )

        self.__required_operation = op
        return op

    def get_assertions(self) -> List[List[str]]:
        op = self.get_op()
        if op != op.INJECT:
            return []
        host_snap_repo = self._get_snap_repo()
        host_snap_info = host_snap_repo.get_local_snap_info()

        if host_snap_info["revision"].startswith("x"):
            return []

        assertions = []  # type: List[List[str]]
        assertions.append(
            ["snap-declaration", "snap-name={}".format(host_snap_repo.name)]
        )
        assertions.append(
            [
                "snap-revision",
                "snap-revision={}".format(host_snap_info["revision"]),
                "snap-id={}".format(host_snap_info["id"]),
            ]
        )
        return assertions

    def _set_data(self) -> None:
        op = self.get_op()
        host_snap_repo = self._get_snap_repo()
        install_cmd = ["sudo", "snap"]

        if op == _SnapOp.INJECT:
            install_cmd.append("install")
            host_snap_info = host_snap_repo.get_local_snap_info()
            snap_revision = host_snap_info["revision"]

            if snap_revision.startswith("x"):
                install_cmd.append("--dangerous")

            if host_snap_info["confinement"] == "classic":
                install_cmd.append("--classic")

            snap_file_name = "{}_{}.snap".format(host_snap_repo.name, snap_revision)
            install_cmd.append(os.path.join(self._snap_dir, snap_file_name))
        elif op == _SnapOp.INSTALL or op == _SnapOp.REFRESH:
            install_cmd.append(op.name.lower())
            snap_channel = _get_snap_channel(self.snap_name)
            store_snap_info = storeapi.StoreClient().cpi.get_package(
                self.snap_name, snap_channel, self._snap_arch
            )
            snap_revision = store_snap_info["revision"]
            confinement = store_snap_info["confinement"]
            if confinement == "classic":
                install_cmd.append("--classic")
            install_cmd.extend(["--channel", snap_channel])
            install_cmd.append(host_snap_repo.name)
        elif op == _SnapOp.NOP:
            install_cmd = []
            snap_revision = None

        self.__install_cmd = install_cmd
        self.__revision = snap_revision

    def get_revision(self) -> str:
        if self.__revision is None:
            self._set_data()
        return self.__revision

    def get_install_cmd(self) -> List[str]:
        if self.__install_cmd is None:
            self._set_data()
        return self.__install_cmd


def _load_registry(registry_filepath: Optional[str]) -> Dict[str, List[Any]]:
    if registry_filepath is None or not os.path.exists(registry_filepath):
        return dict()

    with open(registry_filepath) as registry_file:
        return yaml_utils.load(registry_file)


def _save_registry(
    registry_data: Dict[str, List[Any]], registry_filepath: Optional[str]
) -> None:
    if registry_filepath is None:
        return

    dirpath = os.path.dirname(registry_filepath)
    if dirpath:
        os.makedirs(dirpath, exist_ok=True)

    with open(registry_filepath, "w") as registry_file:
        yaml_utils.dump(registry_data, stream=registry_file)


class SnapInjector:
    """Handle the process of adding snaps into the build environment.

    The specific knowledge of the build environment where these snaps will
    be injected is not required, instead runnables to execute the required
    operations against the build environment are provided upon initialization.

    The snaps to install or refresh in the build environment are added by calling
    add and finally applied by calling apply. If an operation is required, the
    revision that eventually made it into the environment is recorded in the
    registry.
    """

    def __init__(
        self,
        *,
        snap_dir: str,
        registry_filepath: str,
        snap_arch: str,
        runner: Callable[..., None],
        snap_dir_mounter: Callable[[], None],
        snap_dir_unmounter: Callable[[], None],
        file_pusher: Callable[..., None],
        inject_from_host: bool = True
    ) -> None:
        """
        Initialize a SnapInjector instance.

        :param str snap_dir: directory where snaps from the host live for the cases
                             where injection of the snap into the build environment
                             is possible.
        :param str snap_arch: the snap architecture of the snaps to inject.
        :param str registry_filepath: path to where recordings of previusly installed
                                      revisions of a snap can be queried and recorded.
        :param runner: a callable which can run commands in the build environment.
        :param snap_dir_mounter: a callable which can mount the directory where snaps live
                                 on the host into the build environment. This callable
                                 must mount into snap_dir.
        :param snap_dir_unmounter: a callable which can unmount snap_dir from the environment.
        :param file_pusher: a callable that can push file from the host into the build
                            environment.
        :param bool inject_from_host: whether to look for snaps on the host and inject them.
        """

        self._snaps = []  # type: List[_SnapManager]
        self._snap_dir = snap_dir
        self._registry_filepath = registry_filepath
        self._snap_arch = snap_arch
        self._inject_from_host = inject_from_host
        self._runner = runner
        self._snap_dir_mounter = snap_dir_mounter
        self._snap_dir_unmounter = snap_dir_unmounter
        self._file_pusher = file_pusher

        self._registry_data = _load_registry(registry_filepath)

    @contextlib.contextmanager
    def _mounted_dir(self) -> Generator:
        if any((s.get_op() == _SnapOp.INJECT for s in self._snaps)):
            self._snap_dir_mounter()
            try:
                yield
            finally:
                self._snap_dir_unmounter()
        else:
            yield

    def _disable_and_wait_for_refreshes(self) -> None:
        # Disable autorefresh for 15 minutes,
        # https://github.com/snapcore/snapd/pull/5436/files
        now_plus_15 = datetime.datetime.now() + datetime.timedelta(minutes=15)
        self._runner(
            [
                "sudo",
                "snap",
                "set",
                "core",
                "refresh.hold={}Z".format(now_plus_15.isoformat()),
            ]
        )
        # Auto refresh may have kicked in while setting the hold.
        logger.debug("Waiting for pending snap auto refreshes.")
        with contextlib.suppress(errors.ProviderExecError):
            self._runner(
                ["sudo", "snap", "watch", "--last=auto-refresh"], hide_output=True
            )

    def _inject_assertions(self) -> None:
        assertions = []  # type: List[List[str]]
        for snap in self._snaps:
            assertions.extend(snap.get_assertions())
        if assertions:
            assertions = [_STORE_ASSERTION] + assertions

        # Return early if not assertions need acking
        if not assertions:
            return

        with tempfile.NamedTemporaryFile() as assertion_file:
            for assertion in assertions:
                assertion_file.write(repo.snaps.get_assertion(assertion))
                assertion_file.write(b"\n")
            assertion_file.flush()

            self._file_pusher(
                source=assertion_file.name, destination=assertion_file.name
            )
            self._runner(["sudo", "snap", "ack", assertion_file.name])

    def _get_latest_revision(self, snap_name) -> Optional[str]:
        try:
            return self._registry_data[snap_name][-1]["revision"]
        except (IndexError, KeyError):
            return None

    def _record_revision(self, snap_name: str, snap_revision: str) -> None:
        entry = dict(revision=snap_revision)

        if snap_name not in self._registry_data:
            self._registry_data[snap_name] = [entry]
        else:
            self._registry_data[snap_name].append(entry)

    def add(self, snap_name: str) -> None:
        self._snaps.append(
            _SnapManager(
                snap_name=snap_name,
                snap_dir=self._snap_dir,
                latest_revision=self._get_latest_revision(snap_name),
                snap_arch=self._snap_arch,
                inject_from_host=self._inject_from_host,
            )
        )

    def apply(self) -> None:
        if all((s.get_op() == _SnapOp.NOP for s in self._snaps)):
            return

        # Disable refreshes so they do not interfere with installation ops.
        self._disable_and_wait_for_refreshes()

        # Before injecting any snap we must make sure the required assertions
        # are acked.
        self._inject_assertions()

        # Filter out snaps with no operations
        snaps = [snap for snap in self._snaps if snap.get_op() != _SnapOp.NOP]

        with self._mounted_dir():
            for snap in snaps:
                install_cmd = snap.get_install_cmd()
                self._runner(install_cmd)
                self._record_revision(snap.snap_name, snap.get_revision())

        _save_registry(self._registry_data, self._registry_filepath)



# -*- Mode:Python; indent-tabs-buildnil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import snapcraft.extractors
from snapcraft.internal.states._state import PartState


def _schema_properties():
    return {
        "after",
        "build-attributes",
        "build-packages",
        "disable-parallel",
        "organize",
        "override-build",
    }


class BuildState(PartState):
    yaml_tag = u"!BuildState"

    def __init__(
        self,
        property_names,
        part_properties=None,
        project=None,
        plugin_assets=None,
        machine_assets=None,
        metadata=None,
        metadata_files=None,
        scriptlet_metadata=None,
    ):
        # Save this off before calling super() since we'll need it
        # FIXME: for 3.x the name `schema_properties` is leaking
        #        implementation details from a higher layer.
        self.schema_properties = property_names
        if plugin_assets:
            self.assets = plugin_assets
        else:
            self.assets = {}
        if machine_assets:
            self.assets.update(machine_assets)

        if not scriptlet_metadata:
            scriptlet_metadata = snapcraft.extractors.ExtractedMetadata()

        if not metadata:
            metadata = snapcraft.extractors.ExtractedMetadata()

        if not metadata_files:
            metadata_files = []

        self.extracted_metadata = {"metadata": metadata, "files": metadata_files}

        self.scriptlet_metadata = scriptlet_metadata

        super().__init__(part_properties, project)

    def properties_of_interest(self, part_properties):
        """Extract the properties concerning this step from part_properties."""

        properties = {}
        for name in self.schema_properties:
            properties[name] = part_properties.get(name)

        for name in _schema_properties():
            properties[name] = part_properties.get(name)

        return properties

    def project_options_of_interest(self, project):
        """Extract the options concerning this step from the project.

        The build step only cares about the target architecture.
        """

        return {"deb_arch": getattr(project, "deb_arch", None)}



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The kernel plugin refines the generic kbuild plugin to allow building
kernel snaps with all the bells and whistles in one shot...

WARNING: this plugin's API is unstable. The cross compiling support is
         experimental.

The following kernel specific options are provided by this plugin:

    - kernel-image-target:
      (yaml object, string or null for default target)
      the default target is bzImage and can be set to any specific
      target.
      For more complex cases where one would want to use
      the same snapcraft.yaml to target multiple architectures a
      yaml object can be used. This yaml object would be a map of
      debian architecture and kernel image build targets.

    - kernel-initrd-modules:
      (array of string)
      list of modules to include in initrd; note that kernel snaps do not
      provide the core boot logic which comes from snappy Ubuntu Core
      OS snap. Include all modules you need for mounting rootfs here.

    - kernel-with-firmware:
      (boolean; default: True)
      use this flag to disable shipping binary firmwares

    - kernel-initrd-firmware:
      (array of string)
      list of firmware files to include in the initrd; these need to be
      relative paths to .installdir and this option does not work if you
      disable building firmware

    - kernel-initrd-compression:
      (string; default: gz)
      initrd compression to use; the only supported value now is 'gz'.

    - kernel-device-trees:
      (array of string)
      list of device trees to build, the format is <device-tree-name>.dts.
"""

import glob
import logging
import os
import shutil
import subprocess
import tempfile

import snapcraft
from snapcraft.plugins import kbuild

logger = logging.getLogger(__name__)


_compression_command = {"gz": "gzip"}

default_kernel_image_target = {
    "amd64": "bzImage",
    "i386": "bzImage",
    "armhf": "zImage",
    "arm64": "Image.gz",
    "powerpc": "uImage",
    "ppc64el": "vmlinux.strip",
    "s390x": "bzImage",
}

required_generic = [
    "DEVTMPFS",
    "DEVTMPFS_MOUNT",
    "TMPFS_POSIX_ACL",
    "IPV6",
    "SYSVIPC",
    "SYSVIPC_SYSCTL",
    "VFAT_FS",
    "NLS_CODEPAGE_437",
    "NLS_ISO8859_1",
]

required_security = [
    "SECURITY",
    "SECURITY_APPARMOR",
    "SYN_COOKIES",
    "STRICT_DEVMEM",
    "DEFAULT_SECURITY_APPARMOR",
    "SECCOMP",
    "SECCOMP_FILTER",
    "CC_STACKPROTECTOR",
    "CC_STACKPROTECTOR_STRONG",
    "DEBUG_RODATA",
    "DEBUG_SET_MODULE_RONX",
]

required_snappy = [
    "RD_LZMA",
    "KEYS",
    "ENCRYPTED_KEYS",
    "SQUASHFS",
    "SQUASHFS_XATTR",
    "SQUASHFS_XZ",
    "DEVPTS_MULTIPLE_INSTANCES",
]

required_systemd = [
    "DEVTMPFS",
    "CGROUPS",
    "INOTIFY_USER",
    "SIGNALFD",
    "TIMERFD",
    "EPOLL",
    "NET",
    "SYSFS",
    "PROC_FS",
    "FHANDLE",
    "BLK_DEV_BSG",
    "NET_NS",
    "IPV6",
    "AUTOFS4_FS",
    "TMPFS_POSIX_ACL",
    "TMPFS_XATTR",
    "SECCOMP",
]

required_boot = ["squashfs"]


class KernelPlugin(kbuild.KBuildPlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()

        schema["properties"]["kernel-image-target"] = {
            "oneOf": [{"type": "string"}, {"type": "object"}],
            "default": "",
        }

        schema["properties"]["kernel-with-firmware"] = {
            "type": "boolean",
            "default": True,
        }

        schema["properties"]["kernel-initrd-modules"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }

        schema["properties"]["kernel-initrd-firmware"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }

        schema["properties"]["kernel-device-trees"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }

        schema["properties"]["kernel-initrd-compression"] = {
            "type": "string",
            "default": "gz",
            "enum": ["gz"],
        }

        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return super().get_build_properties() + [
            "kernel-image-target",
            "kernel-with-firmware",
            "kernel-initrd-modules",
            "kernel-initrd-firmware",
            "kernel-device-trees",
            "kernel-initrd-compression",
        ]

    @property
    def compression_cmd(self):
        return _compression_command[self.options.kernel_initrd_compression]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        # We need to be able to shell out to modprobe
        self.build_packages.append("kmod")

        self._set_kernel_targets()

        self.os_snap = os.path.join(self.sourcedir, "os.snap")
        self.kernel_release = ""

    def enable_cross_compilation(self):
        logger.info(
            "Cross compiling kernel target {!r}".format(self.project.kernel_arch)
        )
        super().enable_cross_compilation()
        # by enabling cross compilation, the kernel_arch and deb_arch
        # from the project options have effectively changed so we reset
        # kernel targets.
        self._set_kernel_targets()

    def _set_kernel_targets(self):
        if not self.options.kernel_image_target:
            self.kernel_image_target = default_kernel_image_target[
                self.project.deb_arch
            ]
        elif isinstance(self.options.kernel_image_target, str):
            self.kernel_image_target = self.options.kernel_image_target
        elif self.project.deb_arch in self.options.kernel_image_target:
            self.kernel_image_target = self.options.kernel_image_target[
                self.project.deb_arch
            ]

        self.make_targets = [self.kernel_image_target, "modules"]
        self.make_install_targets = [
            "modules_install",
            "INSTALL_MOD_PATH={}".format(self.installdir),
        ]
        self.dtbs = ["{}.dtb".format(i) for i in self.options.kernel_device_trees]
        if self.dtbs:
            self.make_targets.extend(self.dtbs)
        elif self.project.kernel_arch == "arm" or self.project.kernel_arch == "arm64":
            self.make_targets.append("dtbs")
            self.make_install_targets.extend(
                ["dtbs_install", "INSTALL_DTBS_PATH={}/dtbs".format(self.installdir)]
            )
        self.make_install_targets.extend(self._get_fw_install_targets())

    def _get_fw_install_targets(self):
        if not self.options.kernel_with_firmware:
            return []

        return [
            "firmware_install",
            "INSTALL_FW_PATH={}".format(
                os.path.join(self.installdir, "lib", "firmware")
            ),
        ]

    def _unpack_generic_initrd(self):
        initrd_path = os.path.join("boot", "initrd.img-core")
        initrd_unpacked_path = os.path.join(self.builddir, "initrd-staging")
        if os.path.exists(initrd_unpacked_path):
            shutil.rmtree(initrd_unpacked_path)
        os.makedirs(initrd_unpacked_path)

        with tempfile.TemporaryDirectory() as temp_dir:
            unsquashfs_path = snapcraft.file_utils.get_tool_path("unsquashfs")
            subprocess.check_call(
                [unsquashfs_path, self.os_snap, os.path.dirname(initrd_path)],
                cwd=temp_dir,
            )

            tmp_initrd_path = os.path.join(temp_dir, "squashfs-root", initrd_path)

            decompressed = False
            # Roll over valid decompression mechanisms until one works
            for decompressor in ("xz", "gzip"):
                try:
                    subprocess.check_call(
                        "cat {0} | {1} -dc | cpio -i".format(
                            tmp_initrd_path, decompressor
                        ),
                        shell=True,
                        cwd=initrd_unpacked_path,
                    )
                except subprocess.CalledProcessError:
                    pass
                else:
                    decompressed = True
                    break

            if not decompressed:
                raise RuntimeError("The initrd file type is unsupported")

        return initrd_unpacked_path

    def _make_initrd(self):

        logger.info(
            "Generating driver initrd for kernel release: {}".format(
                self.kernel_release
            )
        )

        initrd_unpacked_path = self._unpack_generic_initrd()

        modprobe_outs = []

        # modprobe is typically in /sbin, which may not always be available on
        # the PATH. Add it for this call.
        env = os.environ.copy()
        env["PATH"] += ":/sbin"
        for module in self.options.kernel_initrd_modules:
            modprobe_out = self.run_output(
                [
                    "modprobe",
                    "-n",
                    "--show-depends",
                    "-d",
                    self.installdir,
                    "-S",
                    self.kernel_release,
                    module,
                ],
                env=env,
            )
            modprobe_outs.extend(modprobe_out.split(os.linesep))

        modprobe_outs = [_ for _ in modprobe_outs if _]
        modules_path = os.path.join("lib", "modules", self.kernel_release)
        for src in set(_.split()[1] for _ in modprobe_outs):
            dst = os.path.join(
                initrd_unpacked_path, os.path.relpath(src, self.installdir)
            )
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            os.link(src, dst)

        if modprobe_outs:
            for module_info in ["modules.dep", "modules.dep.bin"]:
                module_info_path = os.path.join(modules_path, module_info)
                src = os.path.join(self.installdir, module_info_path)
                dst = os.path.join(initrd_unpacked_path, module_info_path)
                os.link(src, dst)

        # TODO pickup required firmware from modules.
        for firmware in self.options.kernel_initrd_firmware:
            src = os.path.join(self.installdir, firmware)
            dst = os.path.join(initrd_unpacked_path, firmware)
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            if os.path.isdir(src):
                shutil.copytree(src, dst)
            else:
                os.link(src, dst)

        initrd = "initrd-{}.img".format(self.kernel_release)
        initrd_path = os.path.join(self.installdir, initrd)
        subprocess.check_call(
            "find . | cpio --create --format=newc | "
            "{} > {}".format(self.compression_cmd, initrd_path),
            shell=True,
            cwd=initrd_unpacked_path,
        )
        unversioned_initrd_path = os.path.join(self.installdir, "initrd.img")
        os.link(initrd_path, unversioned_initrd_path)

    def _parse_kernel_release(self):
        kernel_release_path = os.path.join(
            self.builddir, "include", "config", "kernel.release"
        )

        with open(kernel_release_path, "r") as f:
            self.kernel_release = f.read().strip()

        if not self.kernel_release:
            raise ValueError(
                "No kernel release version info found at {!r}".format(
                    kernel_release_path
                )
            )

    def _get_build_arch_dir(self):
        return os.path.join(self.builddir, "arch", self.project.kernel_arch, "boot")

    def _copy_vmlinuz(self):
        kernel = "{}-{}".format(self.kernel_image_target, self.kernel_release)
        src = os.path.join(self._get_build_arch_dir(), self.kernel_image_target)
        dst = os.path.join(self.installdir, kernel)
        if not os.path.exists(src):
            raise ValueError(
                "kernel build did not output a vmlinux binary in top level "
                "dir, expected {!r}".format(src)
            )
        os.link(src, dst)
        os.link(src, os.path.join(self.installdir, "kernel.img"))

    def _copy_system_map(self):
        src = os.path.join(self.builddir, "System.map")
        dst = os.path.join(self.installdir, "System.map-{}".format(self.kernel_release))
        if not os.path.exists(src):
            raise ValueError(
                "kernel build did not output a System.map in top level dir"
            )
        os.link(src, dst)

    def _copy_dtbs(self):
        if not self.options.kernel_device_trees:
            return

        dtb_dir = os.path.join(self.installdir, "dtbs")
        os.makedirs(dtb_dir)

        base_path = os.path.join(self._get_build_arch_dir(), "dts")
        for dtb in self.dtbs:
            found_dtbs = glob.glob(os.path.join(base_path, dtb))
            if not found_dtbs:
                raise RuntimeError("No match for dtb {!r} was found".format(dtb))
            for f in found_dtbs:
                os.link(f, os.path.join(dtb_dir, os.path.basename(f)))

    def _do_parse_config(self, config_path):
        builtin = []
        modules = []
        # tokenize .config and store options in builtin[] or modules[]
        with open(config_path) as f:
            for line in f:
                tok = line.strip().split("=")
                items = len(tok)
                if items == 2:
                    opt = tok[0].upper()
                    val = tok[1].upper()
                    if val == "Y":
                        builtin.append(opt)
                    elif val == "M":
                        modules.append(opt)
        return builtin, modules

    def _do_check_config(self, builtin, modules):
        # check the resulting .config has all the necessary options
        msg = (
            "**** WARNING **** WARNING **** WARNING **** WARNING ****\n"
            "Your kernel config is missing some features that Ubuntu Core "
            "recommends or requires.\n"
            "While we will not prevent you from building this kernel snap, "
            "we suggest you take a look at these:\n"
        )
        required_opts = (
            required_generic + required_security + required_snappy + required_systemd
        )
        missing = []

        for code in required_opts:
            opt = "CONFIG_{}".format(code)
            if opt in builtin:
                continue
            elif opt in modules:
                continue
            else:
                missing.append(opt)

        if missing:
            warn = "\n{}\n".format(msg)
            for opt in missing:
                note = ""
                if opt == "CONFIG_CC_STACKPROTECTOR_STRONG":
                    note = "(4.1.x and later versions only)"
                elif opt == "CONFIG_DEVPTS_MULTIPLE_INSTANCES":
                    note = "(4.8.x and earlier versions only)"
                warn += "{} {}\n".format(opt, note)
            logger.warn(warn)

    def _do_check_initrd(self, builtin, modules):
        # check all required_boot[] items are either builtin or part of initrd
        msg = (
            "**** WARNING **** WARNING **** WARNING **** WARNING ****\n"
            "The following features are deemed boot essential for\n"
            "ubuntu core, consider making them static[=Y] or adding\n"
            "the corresponding module to initrd:\n"
        )
        missing = []

        for code in required_boot:
            opt = "CONFIG_{}".format(code.upper())
            if opt in builtin:
                continue
            elif opt in modules:
                if code in self.options.kernel_initrd_modules:
                    continue
                else:
                    missing.append(opt)
            else:
                missing.append(opt)

        if missing:
            warn = "\n{}\n".format(msg)
            for opt in missing:
                warn += "{}\n".format(opt)
            logger.warn(warn)

    def pull(self):
        super().pull()
        snapcraft.download("core", "stable", self.os_snap, self.project.deb_arch)

    def do_configure(self):
        super().do_configure()

        builtin, modules = self._do_parse_config(self.get_config_path())
        self._do_check_config(builtin, modules)
        self._do_check_initrd(builtin, modules)

    def do_install(self):
        super().do_install()

        self._parse_kernel_release()
        self._make_initrd()
        self._copy_vmlinuz()
        self._copy_system_map()
        self._copy_dtbs()

        # upstream kernel installs under $INSTALL_MOD_PATH/lib/modules/
        # but snapd expects modules/ and firmware/
        shutil.move(os.path.join(self.installdir, "lib", "modules"), self.installdir)
        if self.options.kernel_with_firmware:
            shutil.move(
                os.path.join(self.installdir, "lib", "firmware"), self.installdir
            )
        os.rmdir(os.path.join(self.installdir, "lib"))
        # install .config as config-$version
        config = "config-{}".format(self.kernel_release)
        config_path = os.path.join(self.installdir, config)
        dot_config_path = self.get_config_path()
        os.link(dot_config_path, config_path)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import snapcraft.internal.errors


class ProjectLoaderError(snapcraft.internal.errors.SnapcraftError):

    fmt = ""


class InvalidEpochError(ProjectLoaderError):

    fmt = "epochs are positive integers followed by an optional asterisk"


class DuplicateAliasError(ProjectLoaderError):

    fmt = "Multiple parts have the same alias defined: {aliases!r}"

    def __str__(self):
        if isinstance(self.aliases, (list, set)):
            self.aliases = ",".join(self.aliases)

        return super().__str__()


class SnapcraftLogicError(ProjectLoaderError):

    fmt = "Issue detected while analyzing snapcraft.yaml: {message}"

    def __init__(self, message):
        super().__init__(message=message)


class ExtensionBaseRequiredError(ProjectLoaderError):
    fmt = "Extensions can only be used if the snapcraft.yaml specifies a 'base'"


class ExtensionNotFoundError(ProjectLoaderError):
    fmt = (
        "Failed to find extension {extension_name!r}: "
        "a extension by that name does not exist.\n"
        "Check the extension name and try again."
    )

    def __init__(self, extension_name: str) -> None:
        super().__init__(extension_name=extension_name)


class ExtensionPartConflictError(ProjectLoaderError):
    fmt = (
        "Failed to apply extension {extension_name!r}: "
        "this extension adds a part named {part_name!r}, but a part by that name "
        "already exists.\n"
        "Rename the {part_name!r} part to something else and try again."
    )

    def __init__(self, extension_name: str, part_name: str) -> None:
        super().__init__(extension_name=extension_name, part_name=part_name)


class ExtensionUnsupportedBaseError(ProjectLoaderError):
    fmt = (
        "Failed to load extension {extension_name!r}: "
        "this extension does not support the {base!r} base.\n"
        "Either use a different extension, or use a base supported by this extension."
    )

    def __init__(self, extension_name: str, base: str) -> None:
        super().__init__(extension_name=extension_name, base=base)


class ExtensionMissingDocumentationError(ProjectLoaderError):
    fmt = (
        "The {extension_name!r} extension appears to be missing documentation.\n"
        "We would appreciate it if you created a bug report about this at "
        "https://launchpad.net/snapcraft/+filebug"
    )

    def __init__(self, extension_name: str) -> None:
        super().__init__(extension_name=extension_name)


class SnapcraftAfterPartMissingError(ProjectLoaderError):

    fmt = (
        "Failed to get part information: "
        "Cannot find the definition for part {after_part_name!r}, required by part "
        "{part_name!r}.\n"
        "Remote parts are not supported with bases, so make sure that this part is "
        "defined in the `snapcraft.yaml`."
    )

    def __init__(self, part_name, after_part_name):
        super().__init__(part_name=part_name, after_part_name=after_part_name)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The go plugin can be used for go projects using `go get`.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - go-channel:
      (string, default: latest/stable)
      The Snap Store channel to install go from. If set to an empty string,
      go will be installed using the system's traditional package manager.

    - go-packages:
      (list of strings)
      Go packages to fetch, these must be a "main" package. Dependencies
      are pulled in automatically by `go get`.
      Packages that are not "main" will not cause an error, but would
      not be useful either.
      If the package is a part of the go-importpath the local package
      corresponding to those sources will be used.

    - go-importpath:
      (string)
      This entry tells the checked out `source` to live within a certain path
      within `GOPATH`.
      This is not needed and does not affect `go-packages`.

    - go-buildtags:
      (list of strings)
      Tags to use during the go build. Default is not to use any build tags.
"""

import logging
import os
import shutil
from glob import iglob

import snapcraft
from snapcraft import common
from snapcraft.internal import errors


logger = logging.getLogger(__name__)


class GoPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["go-channel"] = {
            "type": "string",
            "default": "latest/stable",
        }
        schema["properties"]["go-packages"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["properties"]["go-importpath"] = {"type": "string", "default": ""}
        schema["properties"]["go-buildtags"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }

        return schema

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return ["go-packages", "go-buildtags"]

    @classmethod
    def get_pull_properties(cls):
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the pull step dirty.
        return ["go-packages"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        self._setup_base_tools(options.go_channel, project.info.base)

        self._gopath = os.path.join(self.partdir, "go")
        self._gopath_src = os.path.join(self._gopath, "src")
        self._gopath_bin = os.path.join(self._gopath, "bin")
        self._gopath_pkg = os.path.join(self._gopath, "pkg")

    def _setup_base_tools(self, go_channel, base):
        if go_channel:
            self.build_snaps.append("go/{}".format(go_channel))
        elif base in ("core16", "core18"):
            self.build_packages.append("golang-go")
        else:
            raise errors.PluginBaseError(part_name=self.name, base=base)

    def pull(self):
        # use -d to only download (build will happen later)
        # use -t to also get the test-deps
        # since we are not using -u the sources will stick to the
        # original checkout.
        super().pull()
        os.makedirs(self._gopath_src, exist_ok=True)

        if any(iglob("{}/**/*.go".format(self.sourcedir), recursive=True)):
            go_package = self._get_local_go_package()
            go_package_path = os.path.join(self._gopath_src, go_package)
            if os.path.islink(go_package_path):
                os.unlink(go_package_path)
            os.makedirs(os.path.dirname(go_package_path), exist_ok=True)
            os.symlink(self.sourcedir, go_package_path)
            self._run(["go", "get", "-t", "-d", "./{}/...".format(go_package)])

        for go_package in self.options.go_packages:
            self._run(["go", "get", "-t", "-d", go_package])

    def clean_pull(self):
        super().clean_pull()

        # Remove the gopath (if present)
        if os.path.exists(self._gopath):
            shutil.rmtree(self._gopath)

    def _get_local_go_package(self):
        if self.options.go_importpath:
            go_package = self.options.go_importpath
        else:
            logger.warning(
                "Please consider setting `go-importpath` for the {!r} "
                "part".format(self.name)
            )
            go_package = os.path.basename(os.path.abspath(self.options.source))
        return go_package

    def _get_local_main_packages(self):
        search_path = "./{}/...".format(self._get_local_go_package())
        packages = self._run_output(
            ["go", "list", "-f", "{{.ImportPath}} {{.Name}}", search_path]
        )
        packages_split = [p.split() for p in packages.splitlines()]
        main_packages = [p[0] for p in packages_split if p[1] == "main"]
        return main_packages

    def build(self):
        super().build()

        tags = []
        if self.options.go_buildtags:
            tags = ["-tags={}".format(",".join(self.options.go_buildtags))]

        packages = self.options.go_packages
        if not packages:
            packages = self._get_local_main_packages()
        for package in packages:
            binary = os.path.join(self._gopath_bin, self._binary_name(package))
            self._run(["go", "build", "-o", binary] + tags + [package])

        install_bin_path = os.path.join(self.installdir, "bin")
        os.makedirs(install_bin_path, exist_ok=True)
        for binary in os.listdir(self._gopath_bin):
            binary_path = os.path.join(self._gopath_bin, binary)
            shutil.copy2(binary_path, install_bin_path)

    def _binary_name(self, package):
        package = package.replace("/...", "")
        return package.split("/")[-1]

    def clean_build(self):
        super().clean_build()

        if os.path.isdir(self._gopath_bin):
            shutil.rmtree(self._gopath_bin)

        if os.path.isdir(self._gopath_pkg):
            shutil.rmtree(self._gopath_pkg)

    def _run(self, cmd, **kwargs):
        env = self._build_environment()
        return self.run(cmd, cwd=self._gopath_src, env=env, **kwargs)

    def _run_output(self, cmd, **kwargs):
        env = self._build_environment()
        return self.run_output(cmd, cwd=self._gopath_src, env=env, **kwargs)

    def _build_environment(self):
        env = os.environ.copy()
        env["GOPATH"] = self._gopath
        env["GOBIN"] = self._gopath_bin

        include_paths = []
        for root in [self.installdir, self.project.stage_dir]:
            include_paths.extend(
                common.get_library_paths(root, self.project.arch_triplet)
            )

        flags = common.combine_paths(include_paths, "-L", " ")
        env["CGO_LDFLAGS"] = "{} {} {}".format(
            env.get("CGO_LDFLAGS", ""), flags, env.get("LDFLAGS", "")
        )

        if self.project.is_cross_compiling:
            env["CC"] = "{}-gcc".format(self.project.arch_triplet)
            env["CXX"] = "{}-g++".format(self.project.arch_triplet)
            env["CGO_ENABLED"] = "1"
            # See https://golang.org/doc/install/source#environment
            go_archs = {"armhf": "arm", "i386": "386", "ppc64el": "ppc64le"}
            env["GOARCH"] = go_archs.get(self.project.deb_arch, self.project.deb_arch)
            if self.project.deb_arch == "armhf":
                env["GOARM"] = "7"
        return env

    def enable_cross_compilation(self):
        pass



import hashlib
import os
import urllib.parse
from time import sleep
from typing import Dict, Iterable, List, TextIO, Union

import pymacaroons
import requests

import snapcraft
from snapcraft import config
from snapcraft.internal.indicators import download_requests_stream

from . import logger
from . import _upload
from . import constants
from . import errors

from ._sso_client import SSOClient
from ._snap_index_client import SnapIndexClient
from ._sca_client import SCAClient
from ._up_down_client import UpDownClient


class StoreClient:
    """High-level client for the V2.0 API SCA resources."""

    def __init__(self) -> None:
        super().__init__()
        self.conf = config.Config()
        self.sso = SSOClient(self.conf)
        self.cpi = SnapIndexClient(self.conf)
        self.updown = UpDownClient(self.conf)
        self.sca = SCAClient(self.conf)

    def login(
        self,
        email: str,
        password: str,
        one_time_password: str = None,
        acls: Iterable[str] = None,
        channels: Iterable[str] = None,
        packages: Iterable[Dict[str, str]] = None,
        expires: str = None,
        config_fd: TextIO = None,
        save: bool = True,
    ) -> None:
        """Log in via the Ubuntu One SSO API."""
        if acls is None:
            acls = [
                "package_access",
                "package_manage",
                "package_push",
                "package_register",
                "package_release",
                "package_update",
            ]

        if config_fd:
            self.conf.load(config_fd=config_fd)
        else:
            # Ask the store for the needed capabilities to be associated with
            # the macaroon.
            macaroon = self.sca.get_macaroon(acls, packages, channels, expires)
            caveat_id = self._extract_caveat_id(macaroon)
            unbound_discharge = self.sso.get_unbound_discharge(
                email, password, one_time_password, caveat_id
            )
            # The macaroon has been discharged, save it in the config
            self.conf.set("macaroon", macaroon)
            self.conf.set("unbound_discharge", unbound_discharge)
            self.conf.set("email", email)

        if save:
            self.conf.save()

    def _extract_caveat_id(self, root_macaroon):
        macaroon = pymacaroons.Macaroon.deserialize(root_macaroon)
        # macaroons are all bytes, never strings
        sso_host = urllib.parse.urlparse(self.sso.root_url).netloc
        for caveat in macaroon.caveats:
            if caveat.location == sso_host:
                return caveat.caveat_id
        else:
            raise errors.InvalidCredentialsError("Invalid root macaroon")

    def logout(self):
        self.conf.clear()
        self.conf.save()

    def _refresh_if_necessary(self, func, *args, **kwargs):
        """Make a request, refreshing macaroons if necessary."""
        try:
            return func(*args, **kwargs)
        except errors.StoreMacaroonNeedsRefreshError:
            unbound_discharge = self.sso.refresh_unbound_discharge(
                self.conf.get("unbound_discharge")
            )
            self.conf.set("unbound_discharge", unbound_discharge)
            self.conf.save()
            return func(*args, **kwargs)

    def whoami(self):
        """Return user relevant login information."""
        account_data = {}

        for k in ("email", "account_id"):
            value = self.conf.get(k)
            if not value:
                account_info = self.get_account_information()
                value = account_info.get(k, "unknown")
                self.conf.set(k, value)
                self.conf.save()
            account_data[k] = value

        return account_data

    def acl(self) -> Dict[str, Union[List[str], str]]:
        """Return permissions for the logged-in user."""

        acl_data = {}

        acl_info = self.verify_acl()
        for key in ("snap_ids", "channels", "permissions", "expires"):
            acl_data[key] = acl_info.get(key)

        return acl_data

    def get_snap_name_for_id(self, snap_id: str) -> str:
        declaration_assertion = self.cpi.get_assertion("snap-declaration", snap_id)
        return declaration_assertion["headers"]["snap-name"]

    def verify_acl(self) -> Dict[str, Union[List[str], str]]:
        return self._refresh_if_necessary(self.sca.verify_acl)

    def get_account_information(self):
        return self._refresh_if_necessary(self.sca.get_account_information)

    def register_key(self, account_key_request):
        return self._refresh_if_necessary(self.sca.register_key, account_key_request)

    def register(self, snap_name, is_private=False):
        return self._refresh_if_necessary(
            self.sca.register, snap_name, is_private, constants.DEFAULT_SERIES
        )

    def push_precheck(self, snap_name):
        return self._refresh_if_necessary(self.sca.snap_push_precheck, snap_name)

    def push_snap_build(self, snap_id, snap_build):
        return self._refresh_if_necessary(self.sca.push_snap_build, snap_id, snap_build)

    def upload(
        self,
        snap_name,
        snap_filename,
        delta_format=None,
        source_hash=None,
        target_hash=None,
        delta_hash=None,
    ):
        # FIXME This should be raised by the function that uses the
        # discharge. --elopio -2016-06-20
        if self.conf.get("unbound_discharge") is None:
            raise errors.InvalidCredentialsError(
                "Unbound discharge not in the config file"
            )

        updown_data = _upload.upload_files(snap_filename, self.updown)

        return self._refresh_if_necessary(
            self.sca.snap_push_metadata,
            snap_name,
            updown_data,
            delta_format=delta_format,
            source_hash=source_hash,
            target_hash=target_hash,
            delta_hash=delta_hash,
        )

    def release(self, snap_name, revision, channels):
        return self._refresh_if_necessary(
            self.sca.snap_release, snap_name, revision, channels
        )

    def get_snap_revisions(self, snap_name, series=None, arch=None):
        if series is None:
            series = constants.DEFAULT_SERIES

        account_info = self.get_account_information()
        try:
            snap_id = account_info["snaps"][series][snap_name]["snap-id"]
        except KeyError:
            raise errors.SnapNotFoundError(snap_name, series=series, arch=arch)

        if snap_id is None:
            raise errors.NoSnapIdError(snap_name)

        response = self._refresh_if_necessary(
            self.sca.snap_revisions, snap_id, series, arch
        )

        if not response:
            raise errors.SnapNotFoundError(snap_name, series=series, arch=arch)

        return response

    def get_snap_status(self, snap_name, series=None, arch=None):
        if series is None:
            series = constants.DEFAULT_SERIES

        account_info = self.get_account_information()
        try:
            snap_id = account_info["snaps"][series][snap_name]["snap-id"]
        except KeyError:
            raise errors.SnapNotFoundError(snap_name, series=series, arch=arch)

        if snap_id is None:
            raise errors.NoSnapIdError(snap_name)

        response = self._refresh_if_necessary(
            self.sca.snap_status, snap_id, series, arch
        )

        if not response:
            raise errors.SnapNotFoundError(snap_name, series=series, arch=arch)

        return response

    def close_channels(self, snap_id, channel_names):
        return self._refresh_if_necessary(
            self.sca.close_channels, snap_id, channel_names
        )

    def download(self, snap_name, channel, download_path, arch=None, except_hash=""):
        if arch is None:
            arch = snapcraft.ProjectOptions().deb_arch

        package = self.cpi.get_package(snap_name, channel, arch)
        if package["download_sha3_384"] != except_hash:
            self._download_snap(
                snap_name,
                channel,
                arch,
                download_path,
                # FIXME LP: #1662665
                package["anon_download_url"],
                package["download_sha512"],
            )
        return package["download_sha3_384"]

    def _download_snap(
        self, name, channel, arch, download_path, download_url, expected_sha512
    ):
        if self._is_downloaded(download_path, expected_sha512):
            logger.info("Already downloaded {} at {}".format(name, download_path))
            return
        logger.info("Downloading {}".format(name))

        # we only resume when redirected to our CDN since we use internap's
        # special sauce.
        resume_possible = False
        total_read = 0
        probe_url = requests.head(download_url)
        if probe_url.is_redirect and "internap" in probe_url.headers["Location"]:
            download_url = probe_url.headers["Location"]
            resume_possible = True

        # HttpAdapter cannot help here as this is a stream.
        # LP: #1617765
        not_downloaded = True
        retry_count = 5
        while not_downloaded and retry_count:
            headers = {}
            if resume_possible and os.path.exists(download_path):
                total_read = os.path.getsize(download_path)
                headers["Range"] = "bytes={}-".format(total_read)
            request = self.cpi.get(download_url, headers=headers, stream=True)
            request.raise_for_status()
            redirections = [h.headers["Location"] for h in request.history]
            if redirections:
                logger.debug(
                    "Redirections for {!r}: {}".format(
                        download_url, ", ".join(redirections)
                    )
                )
            try:
                download_requests_stream(request, download_path, total_read=total_read)
                not_downloaded = False
            except requests.exceptions.ChunkedEncodingError as e:
                logger.debug(
                    "Error while downloading: {!r}. "
                    "Retries left to download: {!r}.".format(e, retry_count)
                )
                retry_count -= 1
                if not retry_count:
                    raise e
                sleep(1)

        if self._is_downloaded(download_path, expected_sha512):
            logger.info("Successfully downloaded {} at {}".format(name, download_path))
        else:
            raise errors.SHAMismatchError(download_path, expected_sha512)

    def _is_downloaded(self, path, expected_sha512):
        if not os.path.exists(path):
            return False

        file_sum = hashlib.sha512()
        with open(path, "rb") as f:
            for file_chunk in iter(lambda: f.read(file_sum.block_size * 128), b""):
                file_sum.update(file_chunk)
        return expected_sha512 == file_sum.hexdigest()

    def push_assertion(self, snap_id, assertion, endpoint, force=False):
        return self.sca.push_assertion(snap_id, assertion, endpoint, force)

    def get_assertion(self, snap_id, endpoint):
        return self.sca.get_assertion(snap_id, endpoint)

    def sign_developer_agreement(self, latest_tos_accepted=False):
        return self.sca.sign_developer_agreement(latest_tos_accepted)

    def push_metadata(self, snap_name, metadata, force):
        """Push the metadata to the server."""
        account_info = self.get_account_information()
        series = constants.DEFAULT_SERIES
        try:
            snap_id = account_info["snaps"][series][snap_name]["snap-id"]
        except KeyError:
            raise errors.SnapNotFoundError(snap_name, series=series)

        if snap_id is None:
            raise errors.NoSnapIdError(snap_name)

        return self._refresh_if_necessary(
            self.sca.push_metadata, snap_id, snap_name, metadata, force
        )

    def push_binary_metadata(self, snap_name, metadata, force):
        """Push the binary metadata to the server."""
        account_info = self.get_account_information()
        series = constants.DEFAULT_SERIES
        try:
            snap_id = account_info["snaps"][series][snap_name]["snap-id"]
        except KeyError:
            raise errors.SnapNotFoundError(snap_name, series=series)

        if snap_id is None:
            raise errors.NoSnapIdError(snap_name)

        return self._refresh_if_necessary(
            self.sca.push_binary_metadata, snap_id, snap_name, metadata, force
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2016 Canonical Ltd
# Copyright (C) 2016 Harald Sitter <sitter@kde.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The autotools plugin is used for autotools based parts.

Autotools based projects are the ones that have the usual
`./configure && make && make install` instruction set.

The plugin tries to build using ./configure first, if it is not there
it will run ./autogen and if autogen is not there it will run autoreconf.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

In addition, this plugin uses the following plugin-specific keywords:

    - configflags:
      (list of strings)
      configure flags to pass to the build such as those shown by running
      './configure --help'
    - install-via:
      (enum, 'destdir' or 'prefix')
      Whether to install via DESTDIR or by using --prefix (default is
      'destdir')
"""

import os
import stat

from snapcraft.plugins import make


class AutotoolsPlugin(make.MakePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["configflags"] = {
            "type": "array",
            "minitems": 1,
            "items": {"type": "string"},
            "default": [],
        }

        schema["properties"]["install-via"] = {
            "enum": ["destdir", "prefix"],
            "default": "destdir",
        }

        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return super().get_build_properties() + ["configflags", "install-via"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)
        self.build_packages.extend(["autoconf", "automake", "autopoint", "libtool"])

        if options.install_via == "destdir":
            self.options.make_install_var = "DESTDIR"
        elif options.install_via == "prefix":
            self.options.make_install_var = ""
        else:
            raise RuntimeError(
                'Unsupported installation method: "{}"'.format(options.install_via)
            )

    def enable_cross_compilation(self):
        pass

    def build(self):
        if not os.path.exists(os.path.join(self.builddir, "configure")):
            generated = False
            scripts = ["autogen.sh", "bootstrap"]
            for script in scripts:
                path = os.path.join(self.builddir, script)
                if not os.path.exists(path) or os.path.isdir(path):
                    continue
                # Make sure it's executable
                if not os.access(path, os.X_OK):
                    os.chmod(
                        path,
                        stat.S_IRUSR
                        | stat.S_IWUSR
                        | stat.S_IXUSR
                        | stat.S_IRGRP
                        | stat.S_IWGRP
                        | stat.S_IXGRP
                        | stat.S_IROTH
                        | stat.S_IWOTH
                        | stat.S_IXOTH,
                    )
                self.run(["env", "NOCONFIGURE=1", "./{}".format(script)])
                generated = True
                break
            if not generated:
                self.run(["autoreconf", "-i"])

        configure_command = ["./configure"]

        if self.options.make_install_var:
            # Use an empty prefix since we'll install via DESTDIR
            configure_command.append("--prefix=")
        else:
            configure_command.append("--prefix=" + self.installdir)
        if self.project.is_cross_compiling:
            configure_command.append("--host={}".format(self.project.arch_triplet))

        self.run(configure_command + self.options.configflags)
        self.make()

    def snap_fileset(self):
        fileset = super().snap_fileset()
        # Remove .la files which don't work when they are moved around
        fileset.append("-**/*.la")
        return fileset



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
from collections import OrderedDict

from snapcraft import formatting_utils
from snapcraft.internal.errors import SnapcraftError


# dict of jsonschema validator -> cause pairs. Wish jsonschema just gave us
# better messages.
_VALIDATION_ERROR_CAUSES = {
    "maxLength": "maximum length is {validator_value}",
    "minLength": "minimum length is {validator_value}",
}


class MissingSnapcraftYamlError(SnapcraftError):

    fmt = (
        "Could not find {snapcraft_yaml_file_path}. Are you sure you are "
        "in the right directory?\n"
        "To start a new project, use `snapcraft init`"
    )

    def __init__(self, *, snapcraft_yaml_file_path):
        super().__init__(snapcraft_yaml_file_path=snapcraft_yaml_file_path)


class YamlValidationError(SnapcraftError):

    fmt = "Issues while validating {source}: {message}"

    @classmethod
    def from_validation_error(cls, error, *, source="snapcraft.yaml"):
        """Take a jsonschema.ValidationError and create a SnapcraftSchemaError.

        The validation errors coming from jsonschema are a nightmare. This
        class tries to make them a bit more understandable.
        """

        messages = []

        preamble = _determine_preamble(error)
        cause = _determine_cause(error)
        supplement = _determine_supplemental_info(error)

        if preamble:
            messages.append(preamble)

        if supplement:
            messages.append(error.message)
            messages.append("({})".format(supplement))
        elif cause:
            messages.append(cause)
        else:
            messages.append(error.message)

        return cls(" ".join(messages), source)

    def __init__(self, message, source="snapcraft.yaml"):
        super().__init__(message=message, source=source)


class DuplicateSnapcraftYamlError(SnapcraftError):

    fmt = (
        "Found a {snapcraft_yaml_file_path!r} and a "
        "{other_snapcraft_yaml_file_path!r}.\n"
        "Please remove one and try again."
    )

    def __init__(
        self, *, snapcraft_yaml_file_path: str, other_snapcraft_yaml_file_path: str
    ) -> None:
        super().__init__(
            snapcraft_yaml_file_path=snapcraft_yaml_file_path,
            other_snapcraft_yaml_file_path=other_snapcraft_yaml_file_path,
        )


class SanityCheckError(SnapcraftError):
    pass


class CommandChainWithLegacyAdapterError(SanityCheckError):
    fmt = (
        "The {app_name!r} app is using 'command-chain' with the 'legacy' adapter.\n"
        "'command-chain' is incompatible with the 'legacy' adapter. Either stop using "
        "'command-chain', or use a different adapter. If you haven't specified either, "
        "you may be using an extension that does. Run 'snapcraft expand-extensions' to "
        "see the snapcraft.yaml with all extensions applied."
    )

    def __init__(self, app_name):
        super().__init__(app_name=app_name)


def _determine_preamble(error):
    messages = []
    path = _determine_property_path(error)
    if path:
        messages.append(
            "The '{}' property does not match the required schema:".format(
                "/".join(path)
            )
        )
    return " ".join(messages)


def _determine_cause(error):
    messages = []

    # error.validator_value may contain a custom validation error message.
    # If so, use it instead of the garbage message jsonschema gives us.
    with contextlib.suppress(TypeError, KeyError):
        messages.append(error.validator_value["validation-failure"].format(error))

    # The schema itself may have a custom validation error message. If so,
    # use it as well.
    with contextlib.suppress(AttributeError, TypeError, KeyError):
        key = error
        if (
            error.schema.get("type") == "object"
            and error.validator == "additionalProperties"
        ):
            key = list(error.instance.keys())[0]

        messages.append(error.schema["validation-failure"].format(key))

    # anyOf failures might have usable context... try to improve them a bit
    if error.validator == "anyOf":
        contextual_messages = OrderedDict()  # type: Dict[str, str]
        for contextual_error in error.context:
            key = contextual_error.schema_path.popleft()
            if key not in contextual_messages:
                contextual_messages[key] = []
            message = contextual_error.message
            if message:
                # Sure it starts lower-case (not all messages do)
                contextual_messages[key].append(message[0].lower() + message[1:])

        oneOf_messages = []  # type: List[str]
        for key, value in contextual_messages.items():
            oneOf_messages.append(formatting_utils.humanize_list(value, "and", "{}"))

        messages.append(formatting_utils.humanize_list(oneOf_messages, "or", "{}"))

    return " ".join(messages)


def _determine_supplemental_info(error):
    message = _VALIDATION_ERROR_CAUSES.get(error.validator, "").format(
        validator_value=error.validator_value
    )

    if not message and error.validator == "anyOf":
        message = _interpret_anyOf(error)

    if not message and error.cause:
        message = error.cause

    return message


def _determine_property_path(error):
    path = []
    while error.absolute_path:
        element = error.absolute_path.popleft()
        # assume numbers are indices and use 'xxx[123]' notation.
        if isinstance(element, int):
            path[-1] = "{}[{}]".format(path[-1], element)
        else:
            path.append(str(element))

    return path


def _interpret_anyOf(error):
    """Interpret a validation error caused by the anyOf validator.

    Returns:
        A string containing a (hopefully) helpful validation error message. It
        may be empty.
    """

    usages = []
    try:
        for validator in error.validator_value:
            usages.append(validator["usage"])
    except (TypeError, KeyError):
        return ""

    return "must be one of {}".format(formatting_utils.humanize_list(usages, "or"))


class MultipassMissingNonLinuxError(SanityCheckError):
    fmt = (
        "You need multipass installed to build snaps:\n"
        "https://github.com/CanonicalLtd/multipass/releases"
    )


class MultipassMissingLinuxError(SanityCheckError):
    fmt = (
        "You need multipass installed to build snaps which use the base keyword.\n"
        "Would you like to install it now?"
    )


class SnapMissingLinuxError(SanityCheckError):
    fmt = (
        "You need multipass installed to build snaps which use the base keyword.\n"
        "Enable snap support (https://docs.snapcraft.io/core/install) and run:\n"
        "snap install multipass --classic --beta"
    )
